# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

# Database Module
# --------------------

from __future__ import unicode_literals

import re
import time
import frappe
import datetime
import frappe.defaults
# import frappe.model.meta

from frappe import _
from time import time
from frappe.utils import now, getdate, cast_fieldtype
from frappe.utils.background_jobs import execute_job, get_queue
from frappe.model.utils.link_count import flush_local_link_count
from frappe.utils import cint

# imports - compatibility imports
from six import (
	integer_types,
	string_types,
	text_type,
	iteritems
)

class Database(object):
	"""
	   Open a database connection with the given parmeters, if use_default is True, use the
	   login details from `conf.py`. This is called by the request handler and is accessible using
	   the `db` global variable. the `sql` method is also global to run queries
	"""
	VARCHAR_LEN = 140
	MAX_COLUMN_LENGTH = 64

	OPTIONAL_COLUMNS = ["_user_tags", "_comments", "_assign", "_liked_by"]
	DEFAULT_SHORTCUTS = ['_Login', '__user', '_Full Name', 'Today', '__today', "now", "Now"]
	STANDARD_VARCHAR_COLUMNS = ('name', 'owner', 'modified_by', 'parent', 'parentfield', 'parenttype')
	DEFAULT_COLUMNS = ['name', 'creation', 'modified', 'modified_by', 'owner', 'docstatus', 'parent',
		'parentfield', 'parenttype', 'idx']

	class InvalidColumnName(frappe.ValidationError): pass


	def __init__(self, host=None, user=None, password=None, ac_name=None, use_default=0, port=None):
		self.setup_type_map()
		self.host = host or frappe.conf.db_host or 'localhost'
		self.port = port or frappe.conf.db_port or ''
		self.user = user or frappe.conf.db_name
		self.db_name = frappe.conf.db_name
		self._conn = None

		if ac_name:
			self.user = ac_name or frappe.conf.db_name

		if use_default:
			self.user = frappe.conf.db_name

		self.transaction_writes = 0
		self.auto_commit_on_many_writes = 0

		self.password = password or frappe.conf.db_password
		self.value_cache = {}

	def setup_type_map(self):
		pass

	def connect(self):
		"""Connects to a database as set in `site_config.json`."""
		self.cur_db_name = self.user
		self._conn = self.get_connection()
		self._cursor = self._conn.cursor()
		frappe.local.rollback_observers = []

	def use(self, db_name):
		"""`USE` db_name."""
		self._conn.select_db(db_name)

	def get_connection(self):
		pass

	def get_database_size(self):
		pass

	def sql(self, query, values=(), as_dict = 0, as_list = 0, formatted = 0,
		debug=0, ignore_ddl=0, as_utf8=0, auto_commit=0, update=None, explain=False):
		"""Execute a SQL query and fetch all rows.

		:param query: SQL query.
		:param values: List / dict of values to be escaped and substituted in the query.
		:param as_dict: Return as a dictionary.
		:param as_list: Always return as a list.
		:param formatted: Format values like date etc.
		:param debug: Print query and `EXPLAIN` in debug log.
		:param ignore_ddl: Catch exception if table, column missing.
		:param as_utf8: Encode values as UTF 8.
		:param auto_commit: Commit after executing the query.
		:param update: Update this dict to all rows (if returned `as_dict`).

		Examples:

			# return customer names as dicts
			frappe.db.sql("select name from tabCustomer", as_dict=True)

			# return names beginning with a
			frappe.db.sql("select name from tabCustomer where name like %s", "a%")

			# values as dict
			frappe.db.sql("select name from tabCustomer where name like %(name)s and owner=%(owner)s",
				{"name": "a%", "owner":"test@example.com"})

		"""
		if re.search(r'ifnull\(', query, flags=re.IGNORECASE):
			# replaces ifnull in query with coalesce
			query = re.sub(r'ifnull\(', 'coalesce(', query, flags=re.IGNORECASE)

		if not self._conn:
			self.connect()

		# in transaction validations
		self.check_transaction_status(query)

		# autocommit
		if auto_commit: self.commit()

		# execute
		try:
			if debug:
				time_start = time()

			if values!=():
				if isinstance(values, dict):
					values = dict(values)

				# MySQL-python==1.2.5 hack!
				if not isinstance(values, (dict, tuple, list)):
					values = (values,)

				if debug and query.strip().lower().startswith('select'):
					try:
						if explain:
							self.explain_query(query, values)
						frappe.errprint(query % values)
					except TypeError:
						frappe.errprint([query, values])
				if (frappe.conf.get("logging") or False)==2:
					frappe.log("<<<< query")
					frappe.log(query)
					frappe.log("with values:")
					frappe.log(values)
					frappe.log(">>>>")
				self._cursor.execute(query, values)

				if frappe.flags.in_migrate:
					self.log_touched_tables(query, values)

			else:
				if debug:
					if explain:
						self.explain_query(query)
					frappe.errprint(query)
				if (frappe.conf.get("logging") or False)==2:
					frappe.log("<<<< query")
					frappe.log(query)
					frappe.log(">>>>")

				self._cursor.execute(query)

				if frappe.flags.in_migrate:
					self.log_touched_tables(query)

			if debug:
				time_end = time()
				frappe.errprint(("Execution time: {0} sec").format(round(time_end - time_start, 2)))

		except Exception as e:
			if frappe.conf.db_type == 'postgres':
				self.rollback()

			elif self.is_syntax_error(e):
				# only for mariadb
				frappe.errprint('Syntax error in query:')
				frappe.errprint(query)

			if ignore_ddl and (self.is_missing_column(e) or self.is_missing_table(e) or self.cant_drop_field_or_key(e)):
				pass
			else:
				raise

		if auto_commit: self.commit()

		if not self._cursor.description:
			return ()

		# scrub output if required
		if as_dict:
			ret = self.fetch_as_dict(formatted, as_utf8)
			if update:
				for r in ret:
					r.update(update)
			return ret
		elif as_list:
			return self.convert_to_lists(self._cursor.fetchall(), formatted, as_utf8)
		elif as_utf8:
			return self.convert_to_lists(self._cursor.fetchall(), formatted, as_utf8)
		else:
			return self._cursor.fetchall()

	def explain_query(self, query, values=None):
		"""Print `EXPLAIN` in error log."""
		try:
			frappe.errprint("--- query explain ---")
			if values is None:
				self._cursor.execute("explain " + query)
			else:
				self._cursor.execute("explain " + query, values)
			import json
			frappe.errprint(json.dumps(self.fetch_as_dict(), indent=1))
			frappe.errprint("--- query explain end ---")
		except Exception:
			frappe.errprint("error in query explain")

	def sql_list(self, query, values=(), debug=False):
		"""Return data as list of single elements (first column).

		Example:

			# doctypes = ["DocType", "DocField", "User", ...]
			doctypes = frappe.db.sql_list("select name from DocType")
		"""
		return [r[0] for r in self.sql(query, values, debug=debug)]

	def sql_ddl(self, query, values=(), debug=False):
		"""Commit and execute a query. DDL (Data Definition Language) queries that alter schema
		autocommit in MariaDB."""
		self.commit()
		self.sql(query, debug=debug)

	def check_transaction_status(self, query):
		"""Raises exception if more than 20,000 `INSERT`, `UPDATE` queries are
		executed in one transaction. This is to ensure that writes are always flushed otherwise this
		could cause the system to hang."""
		if self.transaction_writes and \
			query and query.strip().split()[0].lower() in ['start', 'alter', 'drop', 'create', "begin", "truncate"]:
			raise Exception('This statement can cause implicit commit')

		if query and query.strip().lower() in ('commit', 'rollback'):
			self.transaction_writes = 0

		if query[:6].lower() in ('update', 'insert', 'delete'):
			self.transaction_writes += 1
			if self.transaction_writes > 200000:
				if self.auto_commit_on_many_writes:
					self.commit()
				else:
					frappe.throw(_("Too many writes in one request. Please send smaller requests"), frappe.ValidationError)

	def fetch_as_dict(self, formatted=0, as_utf8=0):
		"""Internal. Converts results to dict."""
		result = self._cursor.fetchall()
		ret = []
		if result:
			keys = [column[0] for column in self._cursor.description]

		for r in result:
			values = []
			for value in r:
				if as_utf8 and isinstance(value, text_type):
					value = value.encode('utf-8')
				values.append(value)

			ret.append(frappe._dict(zip(keys, values)))
		return ret

	@staticmethod
	def needs_formatting(result, formatted):
		"""Returns true if the first row in the result has a Date, Datetime, Long Int."""
		if result and result[0]:
			for v in result[0]:
				if isinstance(v, (datetime.date, datetime.timedelta, datetime.datetime, integer_types)):
					return True
				if formatted and isinstance(v, (int, float)):
					return True

		return False

	def get_description(self):
		"""Returns result metadata."""
		return self._cursor.description

	@staticmethod
	def convert_to_lists(res, formatted=0, as_utf8=0):
		"""Convert tuple output to lists (internal)."""
		nres = []
		for r in res:
			nr = []
			for val in r:
				if as_utf8 and isinstance(val, text_type):
					val = val.encode('utf-8')
				nr.append(val)
			nres.append(nr)
		return nres

	@staticmethod
	def build_conditions(filters):
		"""Convert filters sent as dict, lists to SQL conditions. filter's key
		is passed by map function, build conditions like:

		* ifnull(`fieldname`, default_value) = %(fieldname)s
		* `fieldname` [=, !=, >, >=, <, <=] %(fieldname)s
		"""
		conditions = []
		values = {}
		def _build_condition(key):
			"""
				filter's key is passed by map function
				build conditions like:
					* ifnull(`fieldname`, default_value) = %(fieldname)s
					* `fieldname` [=, !=, >, >=, <, <=] %(fieldname)s
			"""
			_operator = "="
			_rhs = " %(" + key + ")s"
			value = filters.get(key)
			values[key] = value
			if isinstance(value, (list, tuple)):
				# value is a tuble like ("!=", 0)
				_operator = value[0]
				values[key] = value[1]
				if isinstance(value[1], (tuple, list)):
					# value is a list in tuple ("in", ("A", "B"))
					inner_list = []
					for i, v in enumerate(value[1]):
						inner_key = "{0}_{1}".format(key, i)
						values[inner_key] = v
						inner_list.append("%({0})s".format(inner_key))

					_rhs = " ({0})".format(", ".join(inner_list))
					del values[key]

			if _operator not in ["=", "!=", ">", ">=", "<", "<=", "like", "in", "not in", "not like"]:
				_operator = "="

			if "[" in key:
				split_key = key.split("[")
				condition = "coalesce(`" + split_key[0] + "`, " + split_key[1][:-1] + ") " \
					+ _operator + _rhs
			else:
				condition = "`" + key + "` " + _operator + _rhs

			conditions.append(condition)

		if isinstance(filters, int):
			# docname is a number, convert to string
			filters = str(filters)

		if isinstance(filters, string_types):
			filters = { "name": filters }

		for f in filters:
			_build_condition(f)

		return " and ".join(conditions), values

	def get(self, doctype, filters=None, as_dict=True, cache=False):
		"""Returns `get_value` with fieldname='*'"""
		return self.get_value(doctype, filters, "*", as_dict=as_dict, cache=cache)

	def get_value(self, doctype, filters=None, fieldname="name", ignore=None, as_dict=False,
		debug=False, order_by=None, cache=False):
		"""Returns a document property or list of properties.

		:param doctype: DocType name.
		:param filters: Filters like `{"x":"y"}` or name of the document. `None` if Single DocType.
		:param fieldname: Column name.
		:param ignore: Don't raise exception if table, column is missing.
		:param as_dict: Return values as dict.
		:param debug: Print query in error log.
		:param order_by: Column to order by

		Example:

			# return first customer starting with a
			frappe.db.get_value("Customer", {"name": ("like a%")})

			# return last login of **User** `test@example.com`
			frappe.db.get_value("User", "test@example.com", "last_login")

			last_login, last_ip = frappe.db.get_value("User", "test@example.com",
				["last_login", "last_ip"])

			# returns default date_format
			frappe.db.get_value("System Settings", None, "date_format")
		"""

		ret = self.get_values(doctype, filters, fieldname, ignore, as_dict, debug,
			order_by, cache=cache)

		return ((len(ret[0]) > 1 or as_dict) and ret[0] or ret[0][0]) if ret else None

	def get_values(self, doctype, filters=None, fieldname="name", ignore=None, as_dict=False,
		debug=False, order_by=None, update=None, cache=False):
		"""Returns multiple document properties.

		:param doctype: DocType name.
		:param filters: Filters like `{"x":"y"}` or name of the document.
		:param fieldname: Column name.
		:param ignore: Don't raise exception if table, column is missing.
		:param as_dict: Return values as dict.
		:param debug: Print query in error log.
		:param order_by: Column to order by

		Example:

			# return first customer starting with a
			customers = frappe.db.get_values("Customer", {"name": ("like a%")})

			# return last login of **User** `test@example.com`
			user = frappe.db.get_values("User", "test@example.com", "*")[0]
		"""
		out = None
		if cache and isinstance(filters, string_types) and \
			(doctype, filters, fieldname) in self.value_cache:
			return self.value_cache[(doctype, filters, fieldname)]

		if not order_by: order_by = 'modified desc'

		if isinstance(filters, list):
			out = self._get_value_for_many_names(doctype, filters, fieldname, debug=debug)

		else:
			fields = fieldname
			if fieldname!="*":
				if isinstance(fieldname, string_types):
					fields = [fieldname]
				else:
					fields = fieldname

			if (filters is not None) and (filters!=doctype or doctype=="DocType"):
				try:
					out = self._get_values_from_table(fields, filters, doctype, as_dict, debug, order_by, update)
				except Exception as e:
					if ignore and (frappe.db.is_missing_column(e) or frappe.db.is_table_missing(e)):
						# table or column not found, return None
						out = None
					elif (not ignore) and frappe.db.is_table_missing(e):
						# table not found, look in singles
						out = self.get_values_from_single(fields, filters, doctype, as_dict, debug, update)
					else:
						raise
			else:
				out = self.get_values_from_single(fields, filters, doctype, as_dict, debug, update)

		if cache and isinstance(filters, string_types):
			self.value_cache[(doctype, filters, fieldname)] = out

		return out

	def get_values_from_single(self, fields, filters, doctype, as_dict=False, debug=False, update=None):
		"""Get values from `tabSingles` (Single DocTypes) (internal).

		:param fields: List of fields,
		:param filters: Filters (dict).
		:param doctype: DocType name.
		"""
		# TODO
		# if not frappe.model.meta.is_single(doctype):
		# 	raise frappe.DoesNotExistError("DocType", doctype)

		if fields=="*" or isinstance(filters, dict):
			# check if single doc matches with filters
			values = self.get_singles_dict(doctype)
			if isinstance(filters, dict):
				for key, value in filters.items():
					if values.get(key) != value:
						return []

			if as_dict:
				return values and [values] or []

			if isinstance(fields, list):
				return [map(values.get, fields)]

		else:
			r = self.sql("""select field, value
				from `tabSingles` where field in (%s) and doctype=%s"""
					% (', '.join(['%s'] * len(fields)), '%s'),
					tuple(fields) + (doctype,), as_dict=False, debug=debug)

			if as_dict:
				if r:
					r = frappe._dict(r)
					if update:
						r.update(update)
					return [r]
				else:
					return []
			else:
				return r and [[i[1] for i in r]] or []

	def get_singles_dict(self, doctype, debug = False):
		"""Get Single DocType as dict.

		:param doctype: DocType of the single object whose value is requested

		Example:

			# Get coulmn and value of the single doctype Accounts Settings
			account_settings = frappe.db.get_singles_dict("Accounts Settings")
		"""
		result = self.sql("""
			SELECT field, value
			FROM   `tabSingles`
			WHERE  doctype = %s
		""", doctype)
		# result = _cast_result(doctype, result)

		dict_  = frappe._dict(result)

		return dict_

	@staticmethod
	def get_all(*args, **kwargs):
		return frappe.get_all(*args, **kwargs)

	@staticmethod
	def get_list(*args, **kwargs):
		return frappe.get_list(*args, **kwargs)

	def get_single_value(self, doctype, fieldname, cache=False):
		"""Get property of Single DocType. Cache locally by default

		:param doctype: DocType of the single object whose value is requested
		:param fieldname: `fieldname` of the property whose value is requested

		Example:

			# Get the default value of the company from the Global Defaults doctype.
			company = frappe.db.get_single_value('Global Defaults', 'default_company')
		"""

		if not doctype in self.value_cache:
			self.value_cache = self.value_cache[doctype] = {}

		if fieldname in self.value_cache[doctype]:
			return self.value_cache[doctype][fieldname]

		val = self.sql("""select `value` from
			`tabSingles` where `doctype`=%s and `field`=%s""", (doctype, fieldname))
		val = val[0][0] if val else None

		df = frappe.get_meta(doctype).get_field(fieldname)

		if not df:
			frappe.throw(_('Invalid field name: {0}').format(frappe.bold(fieldname)), self.InvalidColumnName)

		if df.fieldtype in frappe.model.numeric_fieldtypes:
			val = cint(val)

		self.value_cache[doctype][fieldname] = val

		return val

	def get_singles_value(self, *args, **kwargs):
		"""Alias for get_single_value"""
		return self.get_single_value(*args, **kwargs)

	def _get_values_from_table(self, fields, filters, doctype, as_dict, debug, order_by=None, update=None):
		fl = []
		if isinstance(fields, (list, tuple)):
			for f in fields:
				if "(" in f or " as " in f: # function
					fl.append(f)
				else:
					fl.append("`" + f + "`")
			fl = ", ".join(fl)
		else:
			fl = fields
			if fields=="*":
				as_dict = True

		conditions, values = self.build_conditions(filters)

		order_by = ("order by " + order_by) if order_by else ""

		r = self.sql("select {0} from `tab{1}` {2} {3} {4}"
			.format(fl, doctype, "where" if conditions else "", conditions, order_by), values,
			as_dict=as_dict, debug=debug, update=update)

		return r

	def _get_value_for_many_names(self, doctype, names, field, debug=False):
		names = list(filter(None, names))

		if names:
			return self.get_all(doctype,
				fields=['name', field],
				filters=[['name', 'in', names]],
				debug=debug, as_list=1)
		else:
			return {}

	def update(self, *args, **kwargs):
		"""Update multiple values. Alias for `set_value`."""
		return self.set_value(*args, **kwargs)

	def set_value(self, dt, dn, field, val, modified=None, modified_by=None,
		update_modified=True, debug=False):
		"""Set a single value in the database, do not call the ORM triggers
		but update the modified timestamp (unless specified not to).

		**Warning:** this function will not call Document events and should be avoided in normal cases.

		:param dt: DocType name.
		:param dn: Document name.
		:param field: Property / field name or dictionary of values to be updated
		:param value: Value to be updated.
		:param modified: Use this as the `modified` timestamp.
		:param modified_by: Set this user as `modified_by`.
		:param update_modified: default True. Set as false, if you don't want to update the timestamp.
		:param debug: Print the query in the developer / js console.
		"""
		if not modified:
			modified = now()
		if not modified_by:
			modified_by = frappe.session.user

		to_update = {}
		if update_modified:
			to_update = {"modified": modified, "modified_by": modified_by}

		if isinstance(field, dict):
			to_update.update(field)
		else:
			to_update.update({field: val})

		if dn and dt!=dn:
			# with table
			conditions, values = self.build_conditions(dn)

			values.update(to_update)

			set_values = []
			for key in to_update:
				set_values.append('`{0}`=%({0})s'.format(key))

			self.sql("""update `tab{0}`
				set {1} where {2}""".format(dt, ', '.join(set_values), conditions),
				values, debug=debug)

		else:
			# for singles
			keys = list(to_update)
			self.sql('''
				delete from `tabSingles`
				where field in ({0}) and
					doctype=%s'''.format(', '.join(['%s']*len(keys))),
					list(keys) + [dt], debug=debug)
			for key, value in iteritems(to_update):
				self.sql('''insert into `tabSingles` (doctype, field, value) values (%s, %s, %s)''',
					(dt, key, value), debug=debug)

		if dt in self.value_cache:
			del self.value_cache[dt]

		frappe.clear_document_cache(dt, dn)

	@staticmethod
	def set(doc, field, val):
		"""Set value in document. **Avoid**"""
		doc.db_set(field, val)

	def touch(self, doctype, docname):
		"""Update the modified timestamp of this document."""
		modified = now()
		self.sql("""update `tab{doctype}` set `modified`=%s
			where name=%s""".format(doctype=doctype), (modified, docname))
		return modified

	@staticmethod
	def set_temp(value):
		"""Set a temperory value and return a key."""
		key = frappe.generate_hash()
		frappe.cache().hset("temp", key, value)
		return key

	@staticmethod
	def get_temp(key):
		"""Return the temperory value and delete it."""
		return frappe.cache().hget("temp", key)

	def set_global(self, key, val, user='__global'):
		"""Save a global key value. Global values will be automatically set if they match fieldname."""
		self.set_default(key, val, user)

	def get_global(self, key, user='__global'):
		"""Returns a global key value."""
		return self.get_default(key, user)

	def get_default(self, key, parent="__default"):
		"""Returns default value as a list if multiple or single"""
		d = self.get_defaults(key, parent)
		return isinstance(d, list) and d[0] or d

	@staticmethod
	def set_default(key, val, parent="__default", parenttype=None):
		"""Sets a global / user default value."""
		frappe.defaults.set_default(key, val, parent, parenttype)

	@staticmethod
	def add_default(key, val, parent="__default", parenttype=None):
		"""Append a default value for a key, there can be multiple default values for a particular key."""
		frappe.defaults.add_default(key, val, parent, parenttype)


	@staticmethod
	def get_defaults(key=None, parent="__default"):
		"""Get all defaults"""
		if key:
			defaults = frappe.defaults.get_defaults(parent)
			d = defaults.get(key, None)
			if(not d and key != frappe.scrub(key)):
				d = defaults.get(frappe.scrub(key), None)
			return d
		else:
			return frappe.defaults.get_defaults(parent)

	def begin(self):
		self.sql("START TRANSACTION")

	def commit(self):
		"""Commit current transaction. Calls SQL `COMMIT`."""
		self.sql("commit")

		frappe.local.rollback_observers = []
		self.flush_realtime_log()
		enqueue_jobs_after_commit()
		flush_local_link_count()

	@staticmethod
	def flush_realtime_log():
		for args in frappe.local.realtime_log:
			frappe.realtime.emit_via_redis(*args)

		frappe.local.realtime_log = []

	def rollback(self):
		"""`ROLLBACK` current transaction."""
		self.sql("rollback")
		self.begin()
		for obj in frappe.local.rollback_observers:
			if hasattr(obj, "on_rollback"):
				obj.on_rollback()
		frappe.local.rollback_observers = []

	def field_exists(self, dt, fn):
		"""Return true of field exists."""
		return self.sql("select name from tabDocField where fieldname=%s and parent=%s", (dt, fn))

	def table_exists(self, doctype):
		"""Returns True if table for given doctype exists."""
		return ("tab" + doctype) in self.get_tables()

	def get_tables(self):
		return [d[0] for d in self.sql("select table_name from information_schema.tables where table_schema not in ('pg_catalog', 'information_schema')")]

	def a_row_exists(self, doctype):
		"""Returns True if atleast one row exists."""
		return self.sql("select name from `tab{doctype}` limit 1".format(doctype=doctype))

	def exists(self, dt, dn=None, cache=False):
		"""Returns true if document exists.

		:param dt: DocType name.
		:param dn: Document name or filter dict."""
		if isinstance(dt, string_types):
			if dt!="DocType" and dt==dn:
				return True # single always exists (!)
			try:
				return self.get_value(dt, dn, "name", cache=cache)
			except Exception:
				return None

		elif isinstance(dt, dict) and dt.get('doctype'):
			try:
				conditions = []
				for d in dt:
					if d == 'doctype': continue
					conditions.append([d, '=', dt[d]])
				return self.get_all(dt['doctype'], filters=conditions, as_list=1)
			except Exception:
				return None

	def count(self, dt, filters=None, debug=False, cache=False):
		"""Returns `COUNT(*)` for given DocType and filters."""
		if cache and not filters:
			cache_count = frappe.cache().get_value('doctype:count:{}'.format(dt))
			if cache_count is not None:
				return cache_count
		if filters:
			conditions, filters = self.build_conditions(filters)
			count = self.sql("""select count(*)
				from `tab%s` where %s""" % (dt, conditions), filters, debug=debug)[0][0]
			return count
		else:
			count = self.sql("""select count(*)
				from `tab%s`""" % (dt,))[0][0]

			if cache:
				frappe.cache().set_value('doctype:count:{}'.format(dt), count, expires_in_sec = 86400)

			return count

	@staticmethod
	def format_date(date):
		return getdate(date).strftime("%Y-%m-%d")

	@staticmethod
	def format_datetime(datetime):
		if not datetime:
			return '0001-01-01 00:00:00.000000'

		if isinstance(datetime, frappe.string_types):
			if ':' not in datetime:
				datetime = datetime + ' 00:00:00.000000'
		else:
			datetime = datetime.strftime("%Y-%m-%d %H:%M:%S.%f")

		return datetime

	def get_creation_count(self, doctype, minutes):
		"""Get count of records created in the last x minutes"""
		from frappe.utils import now_datetime
		from dateutil.relativedelta import relativedelta

		return self.sql("""select count(name) from `tab{doctype}`
			where creation >= %s""".format(doctype=doctype),
			now_datetime() - relativedelta(minutes=minutes))[0][0]

	def get_db_table_columns(self, table):
		"""Returns list of column names from given table."""
		return [r[0] for r in self.sql('''
			select column_name
			from information_schema.columns
			where table_name = %s ''', table)]

	def get_table_columns(self, doctype):
		"""Returns list of column names from given doctype."""
		columns = self.get_db_table_columns('tab' + doctype)
		if not columns:
			raise self.TableMissingError
		return columns

	def has_column(self, doctype, column):
		"""Returns True if column exists in database."""
		return column in self.get_table_columns(doctype)

	def get_column_type(self, doctype, column):
		return self.sql('''SELECT column_type FROM INFORMATION_SCHEMA.COLUMNS
			WHERE table_name = 'tab{0}' AND column_name = '{1}' '''.format(doctype, column))[0][0]

	def has_index(self, table_name, index_name):
		pass

	def add_index(self, doctype, fields, index_name=None):
		pass

	def add_unique(self, doctype, fields, constraint_name=None):
		pass

	@staticmethod
	def get_index_name(fields):
		index_name = "_".join(fields) + "_index"
		# remove index length if present e.g. (10) from index name
		index_name = re.sub(r"\s*\([^)]+\)\s*", r"", index_name)
		return index_name

	def get_system_setting(self, key):
		def _load_system_settings():
			return self.get_singles_dict("System Settings")
		return frappe.cache().get_value("system_settings", _load_system_settings).get(key)

	def close(self):
		"""Close database connection."""
		if self._conn:
			# self._cursor.close()
			self._conn.close()
			self._cursor = None
			self._conn = None

	@staticmethod
	def escape(s, percent=True):
		"""Excape quotes and percent in given string."""
		# implemented in specific class
		pass

	@staticmethod
	def is_column_missing(e):
		return frappe.db.is_missing_column(e)

	def get_descendants(self, doctype, name):
		'''Return descendants of the current record'''
		node_location_indexes = self.get_value(doctype, name, ('lft', 'rgt'))
		if node_location_indexes:
			lft, rgt = node_location_indexes
			return self.sql_list('''select name from `tab{doctype}`
				where lft > {lft} and rgt < {rgt}'''.format(doctype=doctype, lft=lft, rgt=rgt))
		else:
			# when document does not exist
			return []

	def is_missing_table_or_column(self, e):
		return self.is_missing_column(e) or self.is_missing_table(e)

	def multisql(self, sql_dict, values=(), **kwargs):
		current_dialect = frappe.db.db_type or 'mariadb'
		query = sql_dict.get(current_dialect)
		return self.sql(query, values, **kwargs)

	def delete(self, doctype, conditions):
		if conditions:
			conditions, values = self.build_conditions(conditions)
			return self.sql("DELETE FROM `tab{doctype}` where {conditions}".format(
				doctype=doctype,
				conditions=conditions
			), values)
		else:
			frappe.throw(_('No conditions provided'))

	def log_touched_tables(self, query, values=None):
		if values:
			query = frappe.safe_decode(self._cursor.mogrify(query, values))
		if query.strip().lower().split()[0] in ('insert', 'delete', 'update', 'alter'):
			# single_word_regex is designed to match following patterns
			# `tabXxx`, tabXxx and "tabXxx"

			# multi_word_regex is designed to match following patterns
			# `tabXxx Xxx` and "tabXxx Xxx"

			# ([`"]?) Captures " or ` at the begining of the table name (if provided)
			# \1 matches the first captured group (quote character) at the end of the table name
			# multi word table name must have surrounding quotes.

			# (tab([A-Z]\w+)( [A-Z]\w+)*) Captures table names that start with "tab"
			# and are continued with multiple words that start with a captital letter
			# e.g. 'tabXxx' or 'tabXxx Xxx' or 'tabXxx Xxx Xxx' and so on

			single_word_regex = r'([`"]?)(tab([A-Z]\w+))\1'
			multi_word_regex = r'([`"])(tab([A-Z]\w+)( [A-Z]\w+)+)\1'
			tables = []
			for regex in (single_word_regex, multi_word_regex):
				tables += [groups[1] for groups in re.findall(regex, query)]

			if frappe.flags.touched_tables is None:
				frappe.flags.touched_tables = set()
			frappe.flags.touched_tables.update(tables)


def enqueue_jobs_after_commit():
	if frappe.flags.enqueue_after_commit and len(frappe.flags.enqueue_after_commit) > 0:
		for job in frappe.flags.enqueue_after_commit:
			q = get_queue(job.get("queue"), is_async=job.get("is_async"))
			q.enqueue_call(execute_job, timeout=job.get("timeout"),
							kwargs=job.get("queue_args"))
		frappe.flags.enqueue_after_commit = []

# Helpers
def _cast_result(doctype, result):
	batch = [ ]

	try:
		for field, value in result:
			df = frappe.get_meta(doctype).get_field(field)
			if df:
				value = cast_fieldtype(df.fieldtype, value)

			batch.append(tuple([field, value]))
	except frappe.exceptions.DoesNotExistError:
		return result

	return tuple(batch)



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals
import frappe, json
from frappe import _
from frappe.utils import cstr
from frappe.model import default_fields, table_fields
from six import string_types


def make_mapped_doc(method, source_name, selected_children=None, args=None):
	'''Returns the mapped document calling the given mapper method.
	Sets selected_children as flags for the `get_mapped_doc` method.

	Called from `open_mapped_doc` from create_new.js'''
	method = frappe.get_attr(method)

	if method not in frappe.whitelisted:
		raise frappe.PermissionError

	if selected_children:
		selected_children = json.loads(selected_children)

	if args:
		frappe.flags.args = frappe._dict(json.loads(args))

	frappe.flags.selected_children = selected_children or None

	return method(source_name)


def map_docs(method, source_names, target_doc):
	'''Returns the mapped document calling the given mapper method
	with each of the given source docs on the target doc'''
	method = frappe.get_attr(method)
	if method not in frappe.whitelisted:
		raise frappe.PermissionError

	for src in json.loads(source_names):
		target_doc = method(src, target_doc)
	return target_doc

def get_mapped_doc(from_doctype, from_docname, table_maps, target_doc=None,
		postprocess=None, ignore_permissions=False, ignore_child_tables=False):

	# main
	if not target_doc:
		target_doc = frappe.new_doc(table_maps[from_doctype]["doctype"])
	elif isinstance(target_doc, string_types):
		target_doc = frappe.get_doc(json.loads(target_doc))

	if not ignore_permissions and not target_doc.has_permission("create"):
		target_doc.raise_no_permission_to("create")

	source_doc = frappe.get_doc(from_doctype, from_docname)

	if not ignore_permissions:
		if not source_doc.has_permission("read"):
			source_doc.raise_no_permission_to("read")

	map_doc(source_doc, target_doc, table_maps[source_doc.doctype])

	row_exists_for_parentfield = {}

	# children
	if not ignore_child_tables:
		for df in source_doc.meta.get_table_fields():
			source_child_doctype = df.options
			table_map = table_maps.get(source_child_doctype)

			# if table_map isn't explicitly specified check if both source and target have the same fieldname and same table options and both of them don't have no_copy
			if not table_map:
				target_df = target_doc.meta.get_field(df.fieldname)
				if target_df:
					target_child_doctype = target_df.options
					if target_df and target_child_doctype==source_child_doctype and not df.no_copy and not target_df.no_copy:
						table_map = {
							"doctype": target_child_doctype
						}

			if table_map:
				for source_d in source_doc.get(df.fieldname):
					if "condition" in table_map:
						if not table_map["condition"](source_d):
							continue

					# if children are selected (checked from UI) for this table type,
					# and this record is not in the selected children, then continue
					if (frappe.flags.selected_children
						and (df.fieldname in frappe.flags.selected_children)
						and source_d.name not in frappe.flags.selected_children[df.fieldname]):
						continue

					target_child_doctype = table_map["doctype"]
					target_parentfield = target_doc.get_parentfield_of_doctype(target_child_doctype)

					# does row exist for a parentfield?
					if target_parentfield not in row_exists_for_parentfield:
						row_exists_for_parentfield[target_parentfield] = (True
							if target_doc.get(target_parentfield) else False)

					if table_map.get("add_if_empty") and \
						row_exists_for_parentfield.get(target_parentfield):
						continue

					if table_map.get("filter") and table_map.get("filter")(source_d):
						continue

					map_child_doc(source_d, target_doc, table_map, source_doc)

	if postprocess:
		postprocess(source_doc, target_doc)

	target_doc.set_onload("load_after_mapping", True)
	return target_doc

def map_doc(source_doc, target_doc, table_map, source_parent=None):
	if table_map.get("validation"):
		for key, condition in table_map["validation"].items():
			if condition[0]=="=":
				if source_doc.get(key) != condition[1]:
					frappe.throw(_("Cannot map because following condition fails: ")
						+ key + "=" + cstr(condition[1]))

	map_fields(source_doc, target_doc, table_map, source_parent)

	if "postprocess" in table_map:
		table_map["postprocess"](source_doc, target_doc, source_parent)

def map_fields(source_doc, target_doc, table_map, source_parent):
	no_copy_fields = set([d.fieldname for d in source_doc.meta.get("fields") if (d.no_copy==1 or d.fieldtype in table_fields)]
		+ [d.fieldname for d in target_doc.meta.get("fields") if (d.no_copy==1 or d.fieldtype in table_fields)]
		+ list(default_fields)
		+ list(table_map.get("field_no_map", [])))

	for df in target_doc.meta.get("fields"):
		if df.fieldname not in no_copy_fields:
			# map same fields
			val = source_doc.get(df.fieldname)
			if val not in (None, ""):
				target_doc.set(df.fieldname, val)

			elif df.fieldtype == "Link":
				if not target_doc.get(df.fieldname):
					# map link fields having options == source doctype
					if df.options == source_doc.doctype:
						target_doc.set(df.fieldname, source_doc.name)

					elif source_parent and df.options == source_parent.doctype:
						target_doc.set(df.fieldname, source_parent.name)

	# map other fields
	field_map = table_map.get("field_map")

	if field_map:
		if isinstance(field_map, dict):
			for source_key, target_key in field_map.items():
				val = source_doc.get(source_key)
				if val not in (None, ""):
					target_doc.set(target_key, val)
		else:
			for fmap in field_map:
				val = source_doc.get(fmap[0])
				if val not in (None, ""):
					target_doc.set(fmap[1], val)

	# map idx
	if source_doc.idx:
		target_doc.idx = source_doc.idx

	# add fetch
	for df in target_doc.meta.get("fields", {"fieldtype": "Link"}):
		if target_doc.get(df.fieldname):
			map_fetch_fields(target_doc, df, no_copy_fields)

def map_fetch_fields(target_doc, df, no_copy_fields):
	linked_doc = None

	# options should be like "link_fieldname.fieldname_in_liked_doc"
	for fetch_df in target_doc.meta.get("fields", {"fetch_from": "^{0}.".format(df.fieldname)}):
		if not (fetch_df.fieldtype == "Read Only" or fetch_df.read_only):
			continue

		if ((not target_doc.get(fetch_df.fieldname) or fetch_df.fieldtype == "Read Only")
			and fetch_df.fieldname not in no_copy_fields):
			source_fieldname = fetch_df.fetch_from.split(".")[1]

			if not linked_doc:
				try:
					linked_doc = frappe.get_doc(df.options, target_doc.get(df.fieldname))
				except:
					return

			val = linked_doc.get(source_fieldname)

			if val not in (None, ""):
				target_doc.set(fetch_df.fieldname, val)

def map_child_doc(source_d, target_parent, table_map, source_parent=None):
	target_child_doctype = table_map["doctype"]
	target_parentfield = target_parent.get_parentfield_of_doctype(target_child_doctype)
	target_d = frappe.new_doc(target_child_doctype, target_parent, target_parentfield)

	map_doc(source_d, target_d, table_map, source_parent)

	target_d.idx = None
	target_parent.append(target_parentfield, target_d)
	return target_d



from __future__ import unicode_literals

import re
import frappe

from frappe import _
from frappe.utils import cstr, cint, flt


class InvalidColumnName(frappe.ValidationError): pass

class DBTable:
	def __init__(self, doctype, meta=None):
		self.doctype = doctype
		self.table_name = 'tab{}'.format(doctype)
		self.meta = meta or frappe.get_meta(doctype)
		self.columns = {}
		self.current_columns = {}

		# lists for change
		self.add_column = []
		self.change_type = []
		self.change_name = []
		self.add_unique = []
		self.add_index = []
		self.drop_index = []
		self.set_default = []

		# load
		self.get_columns_from_docfields()

	def sync(self):
		if self.is_new():
			self.create()
		else:
			self.alter()

	def create(self):
		pass

	def get_column_definitions(self):
		column_list = [] + frappe.db.DEFAULT_COLUMNS
		ret = []
		for k in list(self.columns):
			if k not in column_list:
				d = self.columns[k].get_definition()
				if d:
					ret.append('`'+ k + '` ' + d)
					column_list.append(k)
		return ret

	def get_index_definitions(self):
		ret = []
		for key, col in self.columns.items():
			if (col.set_index
				and not col.unique
				and col.fieldtype in frappe.db.type_map
				and frappe.db.type_map.get(col.fieldtype)[0]
				not in ('text', 'longtext')):
				ret.append('index `' + key + '`(`' + key + '`)')
		return ret

	def get_columns_from_docfields(self):
		"""
			get columns from docfields and custom fields
		"""
		fl = frappe.db.sql("SELECT * FROM `tabDocField` WHERE parent = %s", self.doctype, as_dict = 1)
		lengths = {}
		precisions = {}
		uniques = {}

		# optional fields like _comments
		if not self.meta.istable:
			for fieldname in frappe.db.OPTIONAL_COLUMNS:
				fl.append({
					"fieldname": fieldname,
					"fieldtype": "Text"
				})

			# add _seen column if track_seen
			if getattr(self.meta, 'track_seen', False):
				fl.append({
					'fieldname': '_seen',
					'fieldtype': 'Text'
				})

		if (not frappe.flags.in_install_db
			and (frappe.flags.in_install != "frappe"
			or frappe.flags.ignore_in_install)):
			custom_fl = frappe.db.sql("""
				SELECT * FROM `tabCustom Field`
				WHERE dt = %s AND docstatus < 2
				""", (self.doctype,), as_dict=1)
			if custom_fl: fl += custom_fl

			# apply length, precision and unique from property setters
			for ps in frappe.get_all("Property Setter",
				fields=["field_name", "property", "value"],
				filters={
					"doc_type": self.doctype,
					"doctype_or_field": "DocField",
					"property": ["in", ["precision", "length", "unique"]]
				}):

				if ps.property=="length":
					lengths[ps.field_name] = cint(ps.value)

				elif ps.property=="precision":
					precisions[ps.field_name] = cint(ps.value)

				elif ps.property=="unique":
					uniques[ps.field_name] = cint(ps.value)

		for f in fl:
			self.columns[f['fieldname']] = DbColumn(self,
				f['fieldname'],
				f['fieldtype'],
				lengths.get(f["fieldname"]) or f.get('length'),
				f.get('default'),
				f.get('search_index'),
				f.get('options'),
				uniques.get(f["fieldname"],
				f.get('unique')),
				precisions.get(f['fieldname']) or f.get('precision'))

	def validate(self):
		"""Check if change in varchar length isn't truncating the columns"""
		if self.is_new():
			return

		self.setup_table_columns()

		columns = [frappe._dict({"fieldname": f, "fieldtype": "Data"}) for f in
			frappe.db.STANDARD_VARCHAR_COLUMNS]
		columns += self.columns.values()

		for col in columns:
			if len(col.fieldname) >= 64:
				frappe.throw(_("Fieldname is limited to 64 characters ({0})")
					.format(frappe.bold(col.fieldname)))

			if 'varchar' in frappe.db.type_map.get(col.fieldtype, ()):

				# validate length range
				new_length = cint(col.length) or cint(frappe.db.VARCHAR_LEN)
				if not (1 <= new_length <= 1000):
					frappe.throw(_("Length of {0} should be between 1 and 1000").format(col.fieldname))

				current_col = self.current_columns.get(col.fieldname, {})
				if not current_col:
					continue
				current_type = self.current_columns[col.fieldname]["type"]
				current_length = re.findall(r'varchar\(([\d]+)\)', current_type)
				if not current_length:
					# case when the field is no longer a varchar
					continue
				current_length = current_length[0]
				if cint(current_length) != cint(new_length):
					try:
						# check for truncation
						max_length = frappe.db.sql("""SELECT MAX(CHAR_LENGTH(`{fieldname}`)) FROM `tab{doctype}`"""
							.format(fieldname=col.fieldname, doctype=self.doctype))

					except frappe.db.InternalError as e:
						if frappe.db.is_missing_column(e):
							# Unknown column 'column_name' in 'field list'
							continue
						else:
							raise

					if max_length and max_length[0][0] and max_length[0][0] > new_length:
						if col.fieldname in self.columns:
							self.columns[col.fieldname].length = current_length

						frappe.msgprint(_("""Reverting length to {0} for '{1}' in '{2}';
							Setting the length as {3} will cause truncation of data.""")
							.format(current_length, col.fieldname, self.doctype, new_length))

	def is_new(self):
		return self.table_name not in frappe.db.get_tables()

	def setup_table_columns(self):
		# TODO: figure out a way to get key data
		for c in frappe.db.get_table_columns_description(self.table_name):
			self.current_columns[c.name.lower()] = c

	def alter(self):
		pass


class DbColumn:
	def __init__(self, table, fieldname, fieldtype, length, default,
		set_index, options, unique, precision):
		self.table = table
		self.fieldname = fieldname
		self.fieldtype = fieldtype
		self.length = length
		self.set_index = set_index
		self.default = default
		self.options = options
		self.unique = unique
		self.precision = precision

	def get_definition(self, with_default=1):
		column_def = get_definition(self.fieldtype, precision=self.precision, length=self.length)

		if not column_def:
			return column_def

		if self.fieldtype in ("Check", "Int"):
			default_value = cint(self.default) or 0
			column_def += ' not null default {0}'.format(default_value)

		elif self.fieldtype in ("Currency", "Float", "Percent"):
			default_value = flt(self.default) or 0
			column_def += ' not null default {0}'.format(default_value)

		elif self.default and (self.default not in frappe.db.DEFAULT_SHORTCUTS) \
			and not self.default.startswith(":") and column_def not in ('text', 'longtext'):
			column_def += " default {}".format(frappe.db.escape(self.default))

		if self.unique and (column_def not in ('text', 'longtext')):
			column_def += ' unique'

		return column_def

	def build_for_alter_table(self, current_def):
		column_type = get_definition(self.fieldtype, self.precision, self.length)

		# no columns
		if not column_type:
			return

		# to add?
		if not current_def:
			self.fieldname = validate_column_name(self.fieldname)
			self.table.add_column.append(self)
			return

		# type
		if (current_def['type'] != column_type):
			self.table.change_type.append(self)

		# unique
		if((self.unique and not current_def['unique']) and column_type not in ('text', 'longtext')):
			self.table.add_unique.append(self)

		# default
		if (self.default_changed(current_def)
			and (self.default not in frappe.db.DEFAULT_SHORTCUTS)
			and not cstr(self.default).startswith(":")
			and not (column_type in ['text','longtext'])):
			self.table.set_default.append(self)

		# index should be applied or dropped irrespective of type change
		if ((current_def['index'] and not self.set_index and not self.unique)
			or (current_def['unique'] and not self.unique)):
			# to drop unique you have to drop index
			self.table.drop_index.append(self)

		elif (not current_def['index'] and self.set_index) and not (column_type in ('text', 'longtext')):
			self.table.add_index.append(self)

	def default_changed(self, current_def):
		if "decimal" in current_def['type']:
			return self.default_changed_for_decimal(current_def)
		else:
			cur_default = current_def['default']
			new_default = self.default
			if cur_default == "NULL" or cur_default is None:
				cur_default = None
			else:
				# Strip quotes from default value
				# eg. database returns default value as "'System Manager'"
				cur_default = cur_default.lstrip("'").rstrip("'")

			fieldtype = self.fieldtype
			if fieldtype in ['Int', 'Check']:
				cur_default = cint(cur_default)
				new_default = cint(new_default)
			elif fieldtype in ['Currency', 'Float', 'Percent']:
				cur_default = flt(cur_default)
				new_default = flt(new_default)
			return cur_default != new_default

	def default_changed_for_decimal(self, current_def):
		try:
			if current_def['default'] in ("", None) and self.default in ("", None):
				# both none, empty
				return False

			elif current_def['default'] in ("", None):
				try:
					# check if new default value is valid
					float(self.default)
					return True
				except ValueError:
					return False

			elif self.default in ("", None):
				# new default value is empty
				return True

			else:
				# NOTE float() raise ValueError when "" or None is passed
				return float(current_def['default'])!=float(self.default)
		except TypeError:
			return True

def validate_column_name(n):
	special_characters = re.findall(r"[\W]", n, re.UNICODE)
	if special_characters:
		special_characters = ", ".join('"{0}"'.format(c) for c in special_characters)
		frappe.throw(_("Fieldname {0} cannot have special characters like {1}").format(
			frappe.bold(cstr(n)), special_characters), frappe.db.InvalidColumnName)
	return n

def validate_column_length(fieldname):
	if len(fieldname) > frappe.db.MAX_COLUMN_LENGTH:
		frappe.throw(_("Fieldname is limited to 64 characters ({0})").format(fieldname))

def get_definition(fieldtype, precision=None, length=None):
	d = frappe.db.type_map.get(fieldtype)

	# convert int to long int if the length of the int is greater than 11
	if fieldtype == "Int" and length and length > 11:
		d = frappe.db.type_map.get("Long Int")

	if not d: return

	coltype = d[0]
	size = d[1] if d[1] else None

	if size:
		if fieldtype in ["Float", "Currency", "Percent"] and cint(precision) > 6:
			size = '21,9'

		if coltype == "varchar" and length:
			size = length

	if size is not None:
		coltype = "{coltype}({size})".format(coltype=coltype, size=size)

	return coltype

def add_column(doctype, column_name, fieldtype, precision=None):
	if column_name in frappe.db.get_table_columns(doctype):
		# already exists
		return

	frappe.db.commit()
	frappe.db.sql("alter table `tab%s` add column %s %s" % (doctype,
		column_name, get_definition(fieldtype, precision)))





from __future__ import unicode_literals
# Settings saved per user basis
# such as page_limit, filters, last_view

import frappe, json
from six import iteritems, string_types
from frappe import safe_decode

# dict for mapping the index and index type for the filters of different views
filter_dict = {
	"doctype": 0,
	"docfield": 1,
	"operator": 2,
	"value": 3
}

def get_user_settings(doctype, for_update=False):
	user_settings = frappe.cache().hget('_user_settings',
		'{0}::{1}'.format(doctype, frappe.session.user))

	if user_settings is None:
		user_settings = frappe.db.sql('''select data from `__UserSettings`
			where `user`=%s and `doctype`=%s''', (frappe.session.user, doctype))
		user_settings = user_settings and user_settings[0][0] or '{}'

		if not for_update:
			update_user_settings(doctype, user_settings, True)

	return user_settings or '{}'

def update_user_settings(doctype, user_settings, for_update=False):
	'''update user settings in cache'''

	if for_update:
		current = json.loads(user_settings)
	else:
		current = json.loads(get_user_settings(doctype, for_update = True))

		if isinstance(current, string_types):
			# corrupt due to old code, remove this in a future release
			current = {}

		current.update(user_settings)

	frappe.cache().hset('_user_settings', '{0}::{1}'.format(doctype, frappe.session.user),
		json.dumps(current))

def sync_user_settings():
	'''Sync from cache to database (called asynchronously via the browser)'''
	for key, data in iteritems(frappe.cache().hgetall('_user_settings')):
		key = safe_decode(key)
		doctype, user = key.split('::') # WTF?
		frappe.db.multisql({
			'mariadb': """INSERT INTO `__UserSettings`(`user`, `doctype`, `data`)
				VALUES (%s, %s, %s)
				ON DUPLICATE key UPDATE `data`=%s""",
			'postgres': """INSERT INTO `__UserSettings` (`user`, `doctype`, `data`)
				VALUES (%s, %s, %s)
				ON CONFLICT ("user", "doctype") DO UPDATE SET `data`=%s""",
		}, (user, doctype, data, data), as_dict=1)



def save(doctype, user_settings):
	user_settings = json.loads(user_settings or '{}')
	update_user_settings(doctype, user_settings)
	return user_settings


def get(doctype):
	return get_user_settings(doctype)


def update_user_settings_data(user_setting, fieldname, old, new, condition_fieldname=None, condition_values=None):
	data = user_setting.get("data")
	if data:
		update = False
		data = json.loads(data)
		for view in ['List', 'Gantt', 'Kanban', 'Calendar', 'Image', 'Inbox', 'Report']:
			view_settings = data.get(view)
			if view_settings and view_settings.get("filters"):
				view_filters = view_settings.get("filters")
				for view_filter in view_filters:
					if condition_fieldname and view_filter[filter_dict[condition_fieldname]] != condition_values:
						continue
					if view_filter[filter_dict[fieldname]] == old:
						view_filter[filter_dict[fieldname]] = new
						update = True
		if update:
			frappe.db.sql("update __UserSettings set data=%s where doctype=%s and user=%s",
				(json.dumps(data), user_setting.doctype, user_setting.user))

			# clear that user settings from the redis cache
			frappe.cache().hset('_user_settings', '{0}::{1}'.format(user_setting.doctype,
				user_setting.user), None)



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals, print_function

import frappe
import json
from frappe.model import no_value_fields, table_fields
from frappe.model.utils.user_settings import update_user_settings_data, sync_user_settings


def rename_field(doctype, old_fieldname, new_fieldname):
	"""This functions assumes that doctype is already synced"""

	meta = frappe.get_meta(doctype, cached=False)
	new_field = meta.get_field(new_fieldname)
	if not new_field:
		print("rename_field: " + (new_fieldname) + " not found in " + doctype)
		return

	if new_field.fieldtype in table_fields:
		# change parentfield of table mentioned in options
		frappe.db.sql("""update `tab%s` set parentfield=%s
			where parentfield=%s""" % (new_field.options.split("\n")[0], "%s", "%s"),
			(new_fieldname, old_fieldname))
	elif new_field.fieldtype not in no_value_fields:
		if meta.issingle:
			frappe.db.sql("""update `tabSingles` set field=%s
				where doctype=%s and field=%s""",
				(new_fieldname, doctype, old_fieldname))
		else:
			# copy field value
			frappe.db.sql("""update `tab%s` set `%s`=`%s`""" % \
				(doctype, new_fieldname, old_fieldname))

		update_reports(doctype, old_fieldname, new_fieldname)
		update_users_report_view_settings(doctype, old_fieldname, new_fieldname)

		if new_field.fieldtype == "Password":
			pass


	# update in property setter
	update_property_setters(doctype, old_fieldname, new_fieldname)

	# update in user settings
	update_user_settings(doctype, old_fieldname, new_fieldname)

def update_reports(doctype, old_fieldname, new_fieldname):
	def _get_new_sort_by(report_dict, report, key):
		sort_by = report_dict.get(key) or ""
		if sort_by:
			sort_by = sort_by.split(".")
			if len(sort_by) > 1:
				if sort_by[0]==doctype and sort_by[1]==old_fieldname:
					sort_by = doctype + "." + new_fieldname
					report_dict["updated"] = True
			elif report.ref_doctype == doctype and sort_by[0]==old_fieldname:
				sort_by = doctype + "." + new_fieldname
				report_dict["updated"] = True

			if isinstance(sort_by, list):
				sort_by = '.'.join(sort_by)

		return sort_by

	reports = frappe.db.sql("""select name, ref_doctype, json from tabReport
		where report_type = 'Report Builder' and ifnull(is_standard, 'No') = 'No'
		and json like %s and json like %s""",
		('%%%s%%' % old_fieldname , '%%%s%%' % doctype), as_dict=True)

	for r in reports:
		report_dict = json.loads(r.json)

		# update filters
		new_filters = []
		if report_dict.get("filters"):
			for f in report_dict.get("filters"):
				if f and len(f) > 1 and f[0] == doctype and f[1] == old_fieldname:
					new_filters.append([doctype, new_fieldname, f[2], f[3]])
					report_dict["updated"] = True
				else:
					new_filters.append(f)

		# update columns
		new_columns = []
		if report_dict.get("columns"):
			for c in report_dict.get("columns"):
				if c and len(c) > 1 and c[0] == old_fieldname and c[1] == doctype:
					new_columns.append([new_fieldname, doctype])
					report_dict["updated"] = True
				else:
					new_columns.append(c)

		# update sort by
		new_sort_by = _get_new_sort_by(report_dict, r, "sort_by")
		new_sort_by_next = _get_new_sort_by(report_dict, r, "sort_by_next")

		if report_dict.get("updated"):
			new_val = json.dumps({
				"filters": new_filters,
				"columns": new_columns,
				"sort_by": new_sort_by,
				"sort_order": report_dict.get("sort_order"),
				"sort_by_next": new_sort_by_next,
				"sort_order_next": report_dict.get("sort_order_next")
			})

			frappe.db.sql("""update `tabReport` set `json`=%s where name=%s""", (new_val, r.name))

def update_users_report_view_settings(doctype, ref_fieldname, new_fieldname):
	user_report_cols = frappe.db.sql("""select defkey, defvalue from `tabDefaultValue` where
		defkey like '_list_settings:%'""")
	for key, value in user_report_cols:
		new_columns = []
		columns_modified = False
		for field, field_doctype in json.loads(value):
			if field == ref_fieldname and field_doctype == doctype:
				new_columns.append([new_fieldname, field_doctype])
				columns_modified=True
			else:
				new_columns.append([field, field_doctype])

		if columns_modified:
			frappe.db.sql("""update `tabDefaultValue` set defvalue=%s
				where defkey=%s""" % ('%s', '%s'), (json.dumps(new_columns), key))

def update_property_setters(doctype, old_fieldname, new_fieldname):
	frappe.db.sql("""update `tabProperty Setter` set field_name = %s
		where doc_type=%s and field_name=%s""", (new_fieldname, doctype, old_fieldname))

	frappe.db.sql('''update `tabCustom Field` set insert_after=%s
		where insert_after=%s and dt=%s''', (new_fieldname, old_fieldname, doctype))


def update_user_settings(doctype, old_fieldname, new_fieldname):
	# store the user settings data from the redis to db
	sync_user_settings()

	user_settings = frappe.db.sql(''' select user, doctype, data from `__UserSettings`
		where doctype=%s and data like "%%%s%%"''', (doctype, old_fieldname), as_dict=1)

	for user_setting in user_settings:
		update_user_settings_data(user_setting, "docfield", old_fieldname, new_fieldname)



from __future__ import unicode_literals

import frappe
from frappe import _
from frappe.database.schema import DBTable

class MariaDBTable(DBTable):
	def create(self):
		add_text = ''

		# columns
		column_defs = self.get_column_definitions()
		if column_defs: add_text += ',\n'.join(column_defs) + ',\n'

		# index
		index_defs = self.get_index_definitions()
		if index_defs: add_text += ',\n'.join(index_defs) + ',\n'

		# create table
		frappe.db.sql("""create table `%s` (
			name varchar({varchar_len}) not null primary key,
			creation datetime(6),
			modified datetime(6),
			modified_by varchar({varchar_len}),
			owner varchar({varchar_len}),
			docstatus int(1) not null default '0',
			parent varchar({varchar_len}),
			parentfield varchar({varchar_len}),
			parenttype varchar({varchar_len}),
			idx int(8) not null default '0',
			%sindex parent(parent),
			index modified(modified))
			ENGINE={engine}
			ROW_FORMAT=COMPRESSED
			CHARACTER SET=utf8mb4
			COLLATE=utf8mb4_unicode_ci""".format(varchar_len=frappe.db.VARCHAR_LEN,
				engine=self.meta.get("engine") or 'InnoDB') % (self.table_name, add_text))

	def alter(self):
		for col in self.columns.values():
			col.build_for_alter_table(self.current_columns.get(col.fieldname.lower()))

		add_column_query = []
		modify_column_query = []
		add_index_query = []
		drop_index_query = []

		columns_to_modify =  set(self.change_type + self.add_unique + self.set_default)

		for col in self.add_column:
			add_column_query.append("ADD COLUMN `{}` {}".format(col.fieldname, col.get_definition()))

		for col in columns_to_modify:
			modify_column_query.append("MODIFY `{}` {}".format(col.fieldname, col.get_definition()))

		for col in self.add_index:
			# if index key not exists
			if not frappe.db.sql("SHOW INDEX FROM `%s` WHERE key_name = %s" %
					(self.table_name, '%s'), col.fieldname):
				add_index_query.append("ADD INDEX `{}`(`{}`)".format(col.fieldname, col.fieldname))

		for col in self.drop_index:
			if col.fieldname != 'name': # primary key
				# if index key exists
				if frappe.db.sql("""SHOW INDEX FROM `{0}`
					WHERE key_name=%s
					AND Non_unique=%s""".format(self.table_name), (col.fieldname, col.unique)):
					drop_index_query.append("drop index `{}`".format(col.fieldname))

		try:
			for query_parts in [add_column_query, modify_column_query, add_index_query, drop_index_query]:
				if query_parts:
					query_body = ", ".join(query_parts)
					query = "ALTER TABLE `{}` {}".format(self.table_name, query_body)
					frappe.db.sql(query)

		except Exception as e:
			# sanitize
			if e.args[0]==1060:
				frappe.throw(str(e))
			elif e.args[0]==1062:
				fieldname = str(e).split("'")[-2]
				frappe.throw(_("{0} field cannot be set as unique in {1}, as there are non-unique existing values".format(
					fieldname, self.table_name)))
			else:
				raise e



import frappe, subprocess, os
from six.moves import input

def setup_database(force, source_sql, verbose):
	root_conn = get_root_connection()
	root_conn.commit()
	root_conn.sql("DROP DATABASE IF EXISTS `{0}`".format(frappe.conf.db_name))
	root_conn.sql("DROP USER IF EXISTS {0}".format(frappe.conf.db_name))
	root_conn.sql("CREATE DATABASE `{0}`".format(frappe.conf.db_name))
	root_conn.sql("CREATE user {0} password '{1}'".format(frappe.conf.db_name,
		frappe.conf.db_password))
	root_conn.sql("GRANT ALL PRIVILEGES ON DATABASE `{0}` TO {0}".format(frappe.conf.db_name))

	# we can't pass psql password in arguments in postgresql as mysql. So
	# set password connection parameter in environment variable
	subprocess_env = os.environ.copy()
	subprocess_env['PGPASSWORD'] = str(frappe.conf.db_password)
	# bootstrap db
	subprocess.check_output([
		'psql', frappe.conf.db_name, '-h', frappe.conf.db_host, '-U',
		frappe.conf.db_name, '-f',
		os.path.join(os.path.dirname(__file__), 'framework_postgres.sql')
	], env=subprocess_env)

	frappe.connect()

def setup_help_database(help_db_name):
	root_conn = get_root_connection()
	root_conn.sql("DROP DATABASE IF EXISTS `{0}`".format(help_db_name))
	root_conn.sql("DROP USER IF EXISTS {0}".format(help_db_name))
	root_conn.sql("CREATE DATABASE `{0}`".format(help_db_name))
	root_conn.sql("CREATE user {0} password '{1}'".format(help_db_name, help_db_name))
	root_conn.sql("GRANT ALL PRIVILEGES ON DATABASE `{0}` TO {0}".format(help_db_name))

def get_root_connection(root_login=None, root_password=None):
	import getpass
	if not frappe.local.flags.root_connection:
		if not root_login:
			root_login = frappe.conf.get("root_login") or None

		if not root_login:
			root_login = input("Enter postgres super user: ")

		if not root_password:
			root_password = frappe.conf.get("root_password") or None

		if not root_password:
			root_password = getpass.getpass("Postgres super user password: ")

		frappe.local.flags.root_connection = frappe.database.get_db(user=root_login, password=root_password)

	return frappe.local.flags.root_connection



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals
"""
Create a new document with defaults set
"""

import copy
import frappe
import frappe.defaults
from frappe.model import data_fieldtypes
from frappe.utils import nowdate, nowtime, now_datetime

def get_new_doc(doctype, parent_doc = None, parentfield = None, as_dict=False):
	if doctype not in frappe.local.new_doc_templates:
		# cache a copy of new doc as it is called
		# frequently for inserts
		frappe.local.new_doc_templates[doctype] = make_new_doc(doctype)

	doc = copy.deepcopy(frappe.local.new_doc_templates[doctype])

	# doc = make_new_doc(doctype)

	set_dynamic_default_values(doc, parent_doc, parentfield)

	if as_dict:
		return doc
	else:
		return frappe.get_doc(doc)

def make_new_doc(doctype):
	doc = frappe.get_doc({
		"doctype": doctype,
		"__islocal": 1,
		"owner": frappe.session.user,
		"docstatus": 0
	})

	set_user_and_static_default_values(doc)

	doc._fix_numeric_types()
	doc = doc.get_valid_dict(sanitize=False)
	doc["doctype"] = doctype
	doc["__islocal"] = 1

	return doc

def set_user_and_static_default_values(doc):
	user_permissions = []
	defaults = frappe.defaults.get_defaults()

	for df in doc.meta.get("fields"):
		if df.fieldtype in data_fieldtypes:
			# user permissions for link options
			doctype_user_permissions = user_permissions.get(df.options, [])
			# Allowed records for the reference doctype (link field) along with default doc
			allowed_records, default_doc = [],[]

			user_default_value = get_user_default_value(df, defaults, doctype_user_permissions, allowed_records, default_doc)
			if user_default_value is not None:
				doc.set(df.fieldname, user_default_value)

			else:
				if df.fieldname != doc.meta.title_field:
					static_default_value = get_static_default_value(df, doctype_user_permissions, allowed_records)
					if static_default_value is not None:
						doc.set(df.fieldname, static_default_value)

def get_user_default_value(df, defaults, doctype_user_permissions, allowed_records, default_doc):
	# don't set defaults for "User" link field using User Permissions!
	if df.fieldtype == "Link" and df.options != "User":
		# 1 - look in user permissions only for document_type==Setup
		# We don't want to include permissions of transactions to be used for defaults.
		if (frappe.get_meta(df.options).document_type=="Setup"
			and not df.ignore_user_permissions and default_doc):
				return default_doc

		# 2 - Look in user defaults
		user_default = defaults.get(df.fieldname)
		is_allowed_user_default = user_default and (not user_permissions_exist(df, doctype_user_permissions)
			or user_default in allowed_records)

		# is this user default also allowed as per user permissions?
		if is_allowed_user_default:
			return user_default

def get_static_default_value(df, doctype_user_permissions, allowed_records):
	# 3 - look in default of docfield
	if df.get("default"):
		if df.default == "__user":
			return frappe.session.user

		elif df.default == "Today":
			return nowdate()

		elif not df.default.startswith(":"):
			# a simple default value
			is_allowed_default_value = (not user_permissions_exist(df, doctype_user_permissions)
				or (df.default in allowed_records))

			if df.fieldtype!="Link" or df.options=="User" or is_allowed_default_value:
				return df.default

	elif (df.fieldtype == "Select" and df.options and df.options not in ("[Select]", "Loading...")):
		return df.options.split("\n")[0]

def set_dynamic_default_values(doc, parent_doc, parentfield):
	# these values should not be cached
	user_permissions = []

	for df in frappe.get_meta(doc["doctype"]).get("fields"):
		if df.get("default"):
			if df.default.startswith(":"):
				default_value = get_default_based_on_another_field(df, user_permissions, parent_doc)
				if default_value is not None and not doc.get(df.fieldname):
					doc[df.fieldname] = default_value

			elif df.fieldtype == "Datetime" and df.default.lower() == "now":
				doc[df.fieldname] = now_datetime()

		if df.fieldtype == "Time":
			doc[df.fieldname] = nowtime()

	if parent_doc:
		doc["parent"] = parent_doc.name
		doc["parenttype"] = parent_doc.doctype

	if parentfield:
		doc["parentfield"] = parentfield

def user_permissions_exist(df, doctype_user_permissions):
	return (df.fieldtype=="Link"
		and not getattr(df, "ignore_user_permissions", False)
		and doctype_user_permissions)

def get_default_based_on_another_field(df, user_permissions, parent_doc):
	# default value based on another document


	ref_doctype =  df.default[1:]
	ref_fieldname = ref_doctype.lower().replace(" ", "_")
	reference_name = parent_doc.get(ref_fieldname) if parent_doc else frappe.db.get_default(ref_fieldname)
	default_value = frappe.db.get_value(ref_doctype, reference_name, df.fieldname)
	is_allowed_default_value = (not user_permissions_exist(df, user_permissions.get(df.options)) or
		(default_value in []))

	# is this allowed as per user permissions
	if is_allowed_default_value:
		return default_value



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

# metadata

'''
Load metadata (DocType) class

Example:

	meta = frappe.get_meta('User')
	if meta.has_field('first_name'):
		print("DocType" table has field "first_name")


'''

from __future__ import unicode_literals, print_function
from datetime import datetime
from six.moves import range
import frappe, json, os
from frappe.utils import cstr, cint
from frappe.model import default_fields, no_value_fields, optional_fields, data_fieldtypes, table_fields
from frappe.model.document import Document
from frappe.model.base_document import load_doctype_module
from frappe import _



def get_meta(doctype, cached=False):
    return load_meta(doctype)


def load_meta(doctype):
    return Meta(doctype)


class Meta(Document):
    _metaclass = True
    default_fields = list(default_fields)[1:]
    special_doctypes = ("DocField", "DocPerm", "Role", "DocType", "Module Def")

    def __init__(self, doctype):
        self._fields = {}
        if isinstance(doctype, dict):
            super(Meta, self).__init__(doctype)

        elif isinstance(doctype, Document):
            super(Meta, self).__init__(doctype.as_dict())
            self.process()

        else:
            super(Meta, self).__init__("DocType", doctype)
            self.process()

    def load_from_db(self):
        try:
            super(Meta, self).load_from_db()
        except frappe.DoesNotExistError:
            if self.doctype == "DocType" and self.name in self.special_doctypes:
                self.__dict__.update(load_doctype_from_file(self.name))
            else:
                raise

    def process(self):
        # don't process for special doctypes
        # prevent's circular dependency
        if self.name in self.special_doctypes:
            return

        self.add_custom_fields()
        self.apply_property_setters()
        self.sort_fields()
        self.get_valid_columns()
        self.set_custom_permissions()

    def as_dict(self, no_nulls=False):
        def serialize(doc):
            out = {}
            for key in doc.__dict__:
                value = doc.__dict__.get(key)

                if isinstance(value, (list, tuple)):
                    if len(value) > 0 and hasattr(value[0], '__dict__'):
                        value = [serialize(d) for d in value]
                    else:
                        # non standard list object, skip
                        continue

                if (isinstance(value, (frappe.text_type, int, float, datetime, list, tuple))
                        or (not no_nulls and value is None)):
                    out[key] = value

            return out

        return serialize(self)

    def get_link_fields(self):
        return self.get("fields", {"fieldtype": "Link", "options": ["!=", "[Select]"]})

    def get_dynamic_link_fields(self):
        if not hasattr(self, '_dynamic_link_fields'):
            self._dynamic_link_fields = self.get("fields", {"fieldtype": "Dynamic Link"})
        return self._dynamic_link_fields

    def get_select_fields(self):
        return self.get("fields", {"fieldtype": "Select", "options": ["not in",
                                                                      ["[Select]", "Loading..."]]})

    def get_image_fields(self):
        return self.get("fields", {"fieldtype": "Attach Image"})

    def get_set_only_once_fields(self):
        '''Return fields with `set_only_once` set'''
        if not hasattr(self, "_set_only_once_fields"):
            self._set_only_once_fields = self.get("fields", {"set_only_once": 1})
        return self._set_only_once_fields

    def get_table_fields(self):
        if not hasattr(self, "_table_fields"):
            if self.name != "DocType":
                self._table_fields = self.get('fields', {"fieldtype": ['in', table_fields]})
            else:
                self._table_fields = doctype_table_fields

        return self._table_fields

    def get_global_search_fields(self):
        '''Returns list of fields with `in_global_search` set and `name` if set'''
        fields = self.get("fields", {"in_global_search": 1, "fieldtype": ["not in", no_value_fields]})
        if getattr(self, 'show_name_in_global_search', None):
            fields.append(frappe._dict(fieldtype='Data', fieldname='name', label='Name'))

        return fields

    def get_valid_columns(self):
        if not hasattr(self, "_valid_columns"):
            if self.name in ("DocType", "DocField", "DocPerm", "Property Setter"):
                self._valid_columns = get_table_columns(self.name)
            else:
                self._valid_columns = self.default_fields + \
                                      [df.fieldname for df in self.get("fields") if df.fieldtype in data_fieldtypes]

        return self._valid_columns

    def get_table_field_doctype(self, fieldname):
        return {"fields": "DocField", "permissions": "DocPerm"}.get(fieldname)

    def get_field(self, fieldname):
        '''Return docfield from meta'''
        if not self._fields:
            for f in self.get("fields"):
                self._fields[f.fieldname] = f

        return self._fields.get(fieldname)

    def has_field(self, fieldname):
        '''Returns True if fieldname exists'''
        return True if self.get_field(fieldname) else False

    def get_label(self, fieldname):
        '''Get label of the given fieldname'''
        df = self.get_field(fieldname)
        if df:
            label = df.label
        else:
            label = {
                        'name': _('ID'),
                        'owner': _('Created By'),
                        'modified_by': _('Modified By'),
                        'creation': _('Created On'),
                        'modified': _('Last Modified On')
                    }.get(fieldname) or _('No Label')
        return label

    def get_options(self, fieldname):
        return self.get_field(fieldname).options

    def get_link_doctype(self, fieldname):
        df = self.get_field(fieldname)

        if df.fieldtype == "Link":
            return df.options

        elif df.fieldtype == "Dynamic Link":
            return self.get_options(df.options)

        else:
            return None

    def get_search_fields(self):
        search_fields = self.search_fields or "name"
        search_fields = [d.strip() for d in search_fields.split(",")]
        if "name" not in search_fields:
            search_fields.append("name")

        return search_fields

    def get_fields_to_fetch(self, link_fieldname=None):
        '''Returns a list of docfield objects for fields whose values
        are to be fetched and updated for a particular link field

        These fields are of type Data, Link, Text, Readonly and their
        fetch_from property is set as `link_fieldname`.`source_fieldname`'''

        out = []

        if not link_fieldname:
            link_fields = [df.fieldname for df in self.get_link_fields()]

        for df in self.fields:
            if df.fieldtype not in no_value_fields and getattr(df, 'fetch_from', None):
                if link_fieldname:
                    if df.fetch_from.startswith(link_fieldname + '.'):
                        out.append(df)
                else:
                    if '.' in df.fetch_from:
                        fieldname = df.fetch_from.split('.', 1)[0]
                        if fieldname in link_fields:
                            out.append(df)

        return out

    def get_list_fields(self):
        list_fields = ["name"] + [d.fieldname \
                                  for d in self.fields if (d.in_list_view and d.fieldtype in data_fieldtypes)]
        if self.title_field and self.title_field not in list_fields:
            list_fields.append(self.title_field)
        return list_fields

    def get_custom_fields(self):
        return [d for d in self.fields if d.get('is_custom_field')]

    def get_title_field(self):
        '''Return the title field of this doctype,
        explict via `title_field`, or `title` or `name`'''
        title_field = getattr(self, 'title_field', None)
        if not title_field and self.has_field('title'):
            title_field = 'title'
        if not title_field:
            title_field = 'name'

        return title_field

    def get_translatable_fields(self):
        '''Return all fields that are translation enabled'''
        return [d.fieldname for d in self.fields if d.translatable]

    def is_translatable(self, fieldname):
        '''Return true of false given a field'''
        field = self.get_field(fieldname)
        return field and field.translatable

    def get_workflow(self):
        pass

    def add_custom_fields(self):
        try:
            self.extend("fields", frappe.db.sql("""SELECT * FROM `tabCustom Field`
				WHERE dt = %s AND docstatus < 2""", (self.name,), as_dict=1,
                                                update={"is_custom_field": 1}))
        except Exception as e:
            if frappe.db.is_table_missing(e):
                return
            else:
                raise

    def apply_property_setters(self):
        property_setters = frappe.db.sql("""select * from `tabProperty Setter` where
			doc_type=%s""", (self.name,), as_dict=1)

        if not property_setters: return

        integer_docfield_properties = [d.fieldname for d in frappe.get_meta('DocField').fields
                                       if d.fieldtype in ('Int', 'Check')]

        for ps in property_setters:
            if ps.doctype_or_field == 'DocType':
                if ps.property_type in ('Int', 'Check'):
                    ps.value = cint(ps.value)

                self.set(ps.property, ps.value)
            else:
                docfield = self.get("fields", {"fieldname": ps.field_name}, limit=1)
                if docfield:
                    docfield = docfield[0]
                else:
                    continue

                if ps.property in integer_docfield_properties:
                    ps.value = cint(ps.value)

                docfield.set(ps.property, ps.value)

    def sort_fields(self):
        """sort on basis of insert_after"""
        custom_fields = sorted(self.get_custom_fields(), key=lambda df: df.idx)

        if custom_fields:
            newlist = []

            # if custom field is at top
            # insert_after is false
            for c in list(custom_fields):
                if not c.insert_after:
                    newlist.append(c)
                    custom_fields.pop(custom_fields.index(c))

            # standard fields
            newlist += [df for df in self.get('fields') if not df.get('is_custom_field')]

            newlist_fieldnames = [df.fieldname for df in newlist]
            for i in range(2):
                for df in list(custom_fields):
                    if df.insert_after in newlist_fieldnames:
                        cf = custom_fields.pop(custom_fields.index(df))
                        idx = newlist_fieldnames.index(df.insert_after)
                        newlist.insert(idx + 1, cf)
                        newlist_fieldnames.insert(idx + 1, cf.fieldname)

                if not custom_fields:
                    break

            # worst case, add remaining custom fields to last
            if custom_fields:
                newlist += custom_fields

            # renum idx
            for i, f in enumerate(newlist):
                f.idx = i + 1

            self.fields = newlist

    def set_custom_permissions(self):
        '''Reset `permissions` with Custom DocPerm if exists'''
        if frappe.flags.in_patch or frappe.flags.in_install:
            return

        if not self.istable and self.name not in ('DocType', 'DocField', 'DocPerm',
                                                  'Custom DocPerm'):
            custom_perms = frappe.get_all('Custom DocPerm', fields='*',
                                          filters=dict(parent=self.name), update=dict(doctype='Custom DocPerm'))
            if custom_perms:
                self.permissions = [Document(d) for d in custom_perms]

    def get_fieldnames_with_value(self):
        return [df.fieldname for df in self.fields if df.fieldtype not in no_value_fields]

    def get_fields_to_check_permissions(self, user_permission_doctypes):
        fields = self.get("fields", {
            "fieldtype": "Link",
            "parent": self.name,
            "ignore_user_permissions": ("!=", 1),
            "options": ("in", user_permission_doctypes)
        })

        if self.name in user_permission_doctypes:
            fields.append(frappe._dict({
                "label": "Name",
                "fieldname": "name",
                "options": self.name
            }))

        return fields

    def get_high_permlevel_fields(self):
        """Build list of fields with high perm level and all the higher perm levels defined."""
        if not hasattr(self, "high_permlevel_fields"):
            self.high_permlevel_fields = []
            for df in self.fields:
                if df.permlevel > 0:
                    self.high_permlevel_fields.append(df)

        return self.high_permlevel_fields

    def get_dashboard_data(self):
        '''Returns dashboard setup related to this doctype.

        This method will return the `data` property in the
        `[doctype]_dashboard.py` file in the doctype folder'''
        data = frappe._dict()
        try:
            module = load_doctype_module(self.name, suffix='_dashboard')
            if hasattr(module, 'get_data'):
                data = frappe._dict(module.get_data())
        except ImportError:
            pass

        return data

    def get_row_template(self):
        return self.get_web_template(suffix='_row')

    def get_web_template(self, suffix=''):
        '''Returns the relative path of the row template for this doctype'''
        module_name = frappe.scrub(self.module)
        doctype = frappe.scrub(self.name)
        template_path = frappe.get_module_path(module_name, 'doctype',
                                               doctype, 'templates', doctype + suffix + '.html')
        if os.path.exists(template_path):
            return '{module_name}/doctype/{doctype_name}/templates/{doctype_name}{suffix}.html'.format(
                module_name=module_name, doctype_name=doctype, suffix=suffix)
        return None

    def is_nested_set(self):
        return self.has_field('lft') and self.has_field('rgt')


def get_table_columns(doctype):
    return frappe.db.get_table_columns(doctype)


def load_doctype_from_file(doctype):
    return dict()


doctype_table_fields = [
    frappe._dict({"fieldname": "fields", "options": "DocField"}),
    frappe._dict({"fieldname": "permissions", "options": "DocPerm"})
]


#######

def is_single(doctype):
    try:
        return frappe.db.get_value("DocType", doctype, "issingle")
    except IndexError:
        raise Exception('Cannot determine whether %s is single' % doctype)


def get_parent_dt(dt):
    parent_dt = frappe.db.get_all('DocField', 'parent', dict(fieldtype=['in', frappe.model.table_fields], options=dt),
                                  limit=1)
    return parent_dt and parent_dt[0].parent or ''


def set_fieldname(field_id, fieldname):
    frappe.db.set_value('DocField', field_id, 'fieldname', fieldname)


def get_field_currency(df, doc=None):
    """get currency based on DocField options and fieldvalue in doc"""
    currency = None

    if not df.get("options"):
        return None

    if not doc:
        return None

    if not getattr(frappe.local, "field_currency", None):
        frappe.local.field_currency = frappe._dict()

    if not (frappe.local.field_currency.get((doc.doctype, doc.name), {}).get(df.fieldname) or
            (doc.parent and frappe.local.field_currency.get((doc.doctype, doc.parent), {}).get(df.fieldname))):

        ref_docname = doc.parent or doc.name

        if ":" in cstr(df.get("options")):
            split_opts = df.get("options").split(":")
            if len(split_opts) == 3 and doc.get(split_opts[1]):
                currency = frappe.get_cached_value(split_opts[0], doc.get(split_opts[1]), split_opts[2])
        else:
            currency = doc.get(df.get("options"))
            if doc.parent:
                if currency:
                    ref_docname = doc.name
                else:
                    currency = frappe.db.get_value(doc.parenttype, doc.parent, df.get("options"))

        if currency:
            frappe.local.field_currency.setdefault((doc.doctype, ref_docname), frappe._dict()) \
                .setdefault(df.fieldname, currency)

    return frappe.local.field_currency.get((doc.doctype, doc.name), {}).get(df.fieldname) or \
           (doc.parent and frappe.local.field_currency.get((doc.doctype, doc.parent), {}).get(df.fieldname))


def get_field_precision(df, doc=None, currency=None):
    """get precision based on DocField options and fieldvalue in doc"""
    from frappe.utils import get_number_format_info

    if cint(df.precision):
        precision = cint(df.precision)

    elif df.fieldtype == "Currency":
        precision = cint(frappe.db.get_default("currency_precision"))
        if not precision:
            number_format = frappe.db.get_default("number_format") or "#,###.##"
            decimal_str, comma_str, precision = get_number_format_info(number_format)
    else:
        precision = cint(frappe.db.get_default("float_precision")) or 3

    return precision


def get_default_df(fieldname):
    if fieldname in default_fields:
        if fieldname in ("creation", "modified"):
            return frappe._dict(
                fieldname=fieldname,
                fieldtype="Datetime"
            )

        else:
            return frappe._dict(
                fieldname=fieldname,
                fieldtype="Data"
            )


def trim_tables(doctype=None):
    """
    Removes database fields that don't exist in the doctype (json or custom field). This may be needed
    as maintenance since removing a field in a DocType doesn't automatically
    delete the db field.
    """
    ignore_fields = default_fields + optional_fields

    filters = {"issingle": 0}
    if doctype:
        filters["name"] = doctype

    for doctype in frappe.db.get_all("DocType", filters=filters):
        doctype = doctype.name
        columns = frappe.db.get_table_columns(doctype)
        fields = frappe.get_meta(doctype).get_fieldnames_with_value()
        columns_to_remove = [f for f in list(set(columns) - set(fields)) if f not in ignore_fields
                             and not f.startswith("_")]
        if columns_to_remove:
            print(doctype, "columns removed:", columns_to_remove)
            columns_to_remove = ", ".join(["drop `{0}`".format(c) for c in columns_to_remove])
            query = """alter table `tab{doctype}` {columns}""".format(
                doctype=doctype, columns=columns_to_remove)
            frappe.db.sql_ddl(query)



import frappe


common_default_keys = ["__default", "__global"]

def set_default(key, value, parent, parenttype="__default"):
    """Override or add a default value.
    Adds default value in table `tabDefaultValue`.

    :param key: Default key.
    :param value: Default value.
    :param parent: Usually, **User** to whom the default belongs.
    :param parenttype: [optional] default is `__default`."""
    if frappe.db.sql('''
		select
			defkey
		from
			`tabDefaultValue`
		where
			defkey=%s and parent=%s
		for update''', (key, parent)):
        frappe.db.sql("""
			delete from
				`tabDefaultValue`
			where
				defkey=%s and parent=%s""", (key, parent))
    if value != None:
        add_default(key, value, parent)
    else:
        _clear_cache(parent)


def add_default(key, value, parent, parenttype=None):
    d = frappe.get_doc({
        "doctype": "DefaultValue",
        "parent": parent,
        "parenttype": parenttype or "__default",
        "parentfield": "system_defaults",
        "defkey": key,
        "defvalue": value
    })
    d.insert(ignore_permissions=True)
    _clear_cache(parent)


def _clear_cache(parent):
    if parent in common_default_keys:
        frappe.clear_cache()
    else:
        frappe.clear_cache(user=parent)



import frappe
from frappe import _
from frappe.utils import cint, flt
from frappe.database.schema import DBTable, get_definition

class PostgresTable(DBTable):
	def create(self):
		add_text = ''

		# columns
		column_defs = self.get_column_definitions()
		if column_defs: add_text += ',\n'.join(column_defs)

		# index
		# index_defs = self.get_index_definitions()
		# TODO: set docstatus length
		# create table
		frappe.db.sql("""create table `%s` (
			name varchar({varchar_len}) not null primary key,
			creation timestamp(6),
			modified timestamp(6),
			modified_by varchar({varchar_len}),
			owner varchar({varchar_len}),
			docstatus smallint not null default '0',
			parent varchar({varchar_len}),
			parentfield varchar({varchar_len}),
			parenttype varchar({varchar_len}),
			idx bigint not null default '0',
			%s)""".format(varchar_len=frappe.db.VARCHAR_LEN) % (self.table_name, add_text))

		frappe.db.commit()

	def alter(self):
		for col in self.columns.values():
			col.build_for_alter_table(self.current_columns.get(col.fieldname.lower()))

		query = []

		for col in self.add_column:
			query.append("ADD COLUMN `{}` {}".format(col.fieldname, col.get_definition()))

		for col in self.change_type:
			using_clause = ""
			if col.fieldtype in ("Datetime"):
				# The USING option of SET DATA TYPE can actually specify any expression
				# involving the old values of the row
				# read more https://www.postgresql.org/docs/9.1/sql-altertable.html
				using_clause = "USING {}::timestamp without time zone".format(col.fieldname)
			elif col.fieldtype in ("Check"):
				using_clause = "USING {}::smallint".format(col.fieldname)

			query.append("ALTER COLUMN {0} TYPE {1} {2}".format(
				col.fieldname,
				get_definition(col.fieldtype, precision=col.precision, length=col.length),
				using_clause)
			)

		for col in self.set_default:
			if col.fieldname=="name":
				continue

			if col.fieldtype in ("Check", "Int"):
				col_default = cint(col.default)

			elif col.fieldtype in ("Currency", "Float", "Percent"):
				col_default = flt(col.default)

			elif not col.default:
				col_default = "NULL"

			else:
				col_default = "{}".format(frappe.db.escape(col.default))

			query.append("ALTER COLUMN `{}` SET DEFAULT {}".format(col.fieldname, col_default))

		create_index_query = ""
		for col in self.add_index:
			# if index key not exists
			create_index_query += 'CREATE INDEX IF NOT EXISTS "{index_name}" ON `{table_name}`(`{field}`);'.format(
				index_name=col.fieldname,
				table_name=self.table_name,
				field=col.fieldname)

		drop_index_query = ""
		for col in self.drop_index:
			# primary key
			if col.fieldname != 'name':
				# if index key exists
				if not frappe.db.has_index(self.table_name, col.fieldname):
					drop_index_query += 'DROP INDEX IF EXISTS "{}" ;'.format(col.fieldname)

		if query:
			try:
				final_alter_query = "ALTER TABLE `{}` {}".format(self.table_name, ", ".join(query))
				if final_alter_query: frappe.db.sql(final_alter_query)
				if create_index_query: frappe.db.sql(create_index_query)
				if drop_index_query: frappe.db.sql(drop_index_query)
			except Exception as e:
				# sanitize
				if frappe.db.is_duplicate_fieldname(e):
					frappe.throw(str(e))
				elif frappe.db.is_duplicate_entry(e):
					fieldname = str(e).split("'")[-2]
					frappe.throw(_("""{0} field cannot be set as unique in {1},
						as there are non-unique existing values""".format(
						fieldname, self.table_name)))
					raise e
				else:
					raise e



import setuptools
import os

# from setuptools import setup

with open('requirements.txt') as f:
    install_requires = f.read().strip().split('\n')

version = '1.0.0'

setuptools.setup(name='frappe_orm',
                 packages=setuptools.find_packages(),
                 version=version,
                 description='Frappe Framework Orm',
                 author='omar hamdy',
                 author_email='omarhamdy49@gmail.com',
                 license='MIT',
                 keywords=["frappe", "orm", "postgres", "mariadb", "database"],
                 zip_safe=False)



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals
from six import iteritems, string_types

import frappe
import datetime
from frappe import _
from frappe.model import default_fields, table_fields
from frappe.model.naming import set_new_name
from frappe.model.utils.link_count import notify_link_count
from frappe.model import display_fieldtypes, data_fieldtypes
from frappe.utils import (cint, flt, now, cstr, getdate, get_datetime, to_timedelta, cast_fieldtype)

max_positive_value = {
    'smallint': 2 ** 15,
    'int': 2 ** 31,
    'bigint': 2 ** 63
}

_classes = {}

doctype_python_modules = {}


def load_doctype_module(doctype, module=None, prefix="", suffix=""):
    """Returns the module object for given doctype."""
    if not module:
        module = get_doctype_module(doctype)

    app = get_module_app(module)

    key = (app, doctype, prefix, suffix)

    module_name = get_module_name(doctype, module, prefix, suffix)

    try:
        if key not in doctype_python_modules:
            doctype_python_modules[key] = frappe.get_module(module_name)
    except ImportError as e:
        raise ImportError('Module import failed for {0} ({1})'.format(doctype, module_name + ' Error: ' + str(e)))

    return doctype_python_modules[key]


def get_doctype_module(doctype):
    """Returns **Module Def** name of given doctype."""

    def make_modules_dict():
        return dict(frappe.db.sql("select name, module from tabDocType"))

    return make_modules_dict()[doctype]


def get_module_name(doctype, module, prefix="", suffix="", app=None):
    return '{app}.{module}.doctype.{doctype}.{prefix}{doctype}{suffix}'.format( \
        app=scrub(app or get_module_app(module)),
        module=scrub(module),
        doctype=scrub(doctype),
        prefix=prefix,
        suffix=suffix)


def scrub(txt):
    return frappe.scrub(txt)


def get_module_app(module):
    try:
        return frappe.local.module_app[scrub(module)]
    except:
        return ""


def get_controller(doctype):
    pass


class BaseDocument(object):
    ignore_in_getter = ("doctype", "_meta", "meta", "_table_fields", "_valid_columns")

    def __init__(self, d):
        self.update(d)
        self.dont_update_if_missing = []

        if hasattr(self, "__setup__"):
            self.__setup__()

    @property
    def meta(self):
        if not hasattr(self, "_meta"):
            self._meta = frappe.get_meta(self.doctype)

        return self._meta

    def update(self, d):
        if "doctype" in d:
            self.set("doctype", d.get("doctype"))

        # first set default field values of base document
        for key in default_fields:
            if key in d:
                self.set(key, d.get(key))

        for key, value in iteritems(d):
            self.set(key, value)

        return self

    def update_if_missing(self, d):
        if isinstance(d, BaseDocument):
            d = d.get_valid_dict()

        if "doctype" in d:
            self.set("doctype", d.get("doctype"))
        for key, value in iteritems(d):
            # dont_update_if_missing is a list of fieldnames, for which, you don't want to set default value
            if (self.get(key) is None) and (value is not None) and (key not in self.dont_update_if_missing):
                self.set(key, value)

    def get_db_value(self, key):
        return frappe.db.get_value(self.doctype, self.name, key)

    def get(self, key=None, filters=None, limit=None, default=None):
        if key:
            if isinstance(key, dict):
                return _filter(self.get_all_children(), key, limit=limit)
            if filters:
                if isinstance(filters, dict):
                    value = _filter(self.__dict__.get(key, []), filters, limit=limit)
                else:
                    default = filters
                    filters = None
                    value = self.__dict__.get(key, default)
            else:
                value = self.__dict__.get(key, default)

            if value is None and key not in self.ignore_in_getter \
                    and key in (d.fieldname for d in self.meta.get_table_fields()):
                self.set(key, [])
                value = self.__dict__.get(key)

            return value
        else:
            return self.__dict__

    def getone(self, key, filters=None):
        return self.get(key, filters=filters, limit=1)[0]

    def set(self, key, value, as_value=False):
        if isinstance(value, list) and not as_value:
            self.__dict__[key] = []
            self.extend(key, value)
        else:
            self.__dict__[key] = value

    def delete_key(self, key):
        if key in self.__dict__:
            del self.__dict__[key]

    def append(self, key, value=None):
        pass

    def extend(self, key, value):
        if isinstance(value, list):
            for v in value:
                self.append(key, v)
        else:
            raise ValueError

    def remove(self, doc):
        self.get(doc.parentfield).remove(doc)

    def _init_child(self, value, key):
        return dict()

    def get_valid_dict(self, sanitize=True, convert_dates_to_str=False, ignore_nulls=False):
        d = frappe._dict()
        for fieldname in self.meta.get_valid_columns():
            d[fieldname] = self.get(fieldname)

            # if no need for sanitization and value is None, continue
            if not sanitize and d[fieldname] is None:
                continue

            df = self.meta.get_field(fieldname)
            if df:
                if df.fieldtype == "Check":
                    d[fieldname] = 1 if cint(d[fieldname]) else 0

                elif df.fieldtype == "Int" and not isinstance(d[fieldname], int):
                    d[fieldname] = cint(d[fieldname])

                elif df.fieldtype in ("Currency", "Float", "Percent") and not isinstance(d[fieldname], float):
                    d[fieldname] = flt(d[fieldname])

                elif df.fieldtype in ("Datetime", "Date", "Time") and d[fieldname] == "":
                    d[fieldname] = None

                elif df.get("unique") and cstr(d[fieldname]).strip() == "":
                    # unique empty field should be set to None
                    d[fieldname] = None

                if isinstance(d[fieldname], list) and df.fieldtype not in table_fields:
                    frappe.throw(_('Value for {0} cannot be a list').format(_(df.label)))

                if convert_dates_to_str and isinstance(d[fieldname],
                                                       (datetime.datetime, datetime.time, datetime.timedelta)):
                    d[fieldname] = str(d[fieldname])

            if d[fieldname] == None and ignore_nulls:
                del d[fieldname]

        return d

    def init_valid_columns(self):
        for key in default_fields:
            if key not in self.__dict__:
                self.__dict__[key] = None

            if key in ("idx", "docstatus") and self.__dict__[key] is None:
                self.__dict__[key] = 0

        for key in self.get_valid_columns():
            if key not in self.__dict__:
                self.__dict__[key] = None

    def get_valid_columns(self):
        if self.doctype not in frappe.local.valid_columns:
            if self.doctype in ("DocField", "DocPerm") and self.parent in ("DocType", "DocField", "DocPerm"):
                from frappe.model.meta import get_table_columns
                valid = get_table_columns(self.doctype)
            else:
                valid = self.meta.get_valid_columns()

            frappe.local.valid_columns[self.doctype] = valid

        return frappe.local.valid_columns[self.doctype]

    def is_new(self):
        return self.get("__islocal")

    def as_dict(self, no_nulls=False, no_default_fields=False, convert_dates_to_str=False):
        doc = self.get_valid_dict(convert_dates_to_str=convert_dates_to_str)
        doc["doctype"] = self.doctype
        for df in self.meta.get_table_fields():
            children = self.get(df.fieldname) or []
            doc[df.fieldname] = [d.as_dict(no_nulls=no_nulls) for d in children]

        if no_nulls:
            for k in list(doc):
                if doc[k] is None:
                    del doc[k]

        if no_default_fields:
            for k in list(doc):
                if k in default_fields:
                    del doc[k]

        for key in ("_user_tags", "__islocal", "__onload", "_liked_by", "__run_link_triggers"):
            if self.get(key):
                doc[key] = self.get(key)

        return doc

    def as_json(self):
        return frappe.as_json(self.as_dict())

    def get_table_field_doctype(self, fieldname):
        return self.meta.get_field(fieldname).options

    def get_parentfield_of_doctype(self, doctype):
        fieldname = [df.fieldname for df in self.meta.get_table_fields() if df.options == doctype]
        return fieldname[0] if fieldname else None

    def db_insert(self):
        """INSERT the document (with valid columns) in the database."""
        if not self.name:
            # name will be set by document class in most cases
            set_new_name(self)

        if not self.creation:
            self.creation = self.modified = now()
            self.created_by = self.modified_by = frappe.session.user

        # if doctype is "DocType", don't insert null values as we don't know who is valid yet
        d = self.get_valid_dict(convert_dates_to_str=True,
                                ignore_nulls=self.doctype in ('DocType', 'DocField', 'DocPerm'))

        columns = list(d)
        try:
            frappe.db.sql("""INSERT INTO `tab{doctype}` ({columns})
					VALUES ({values})""".format(
                doctype=self.doctype,
                columns=", ".join(["`" + c + "`" for c in columns]),
                values=", ".join(["%s"] * len(columns))
            ), list(d.values()))
        except Exception as e:
            if frappe.db.is_primary_key_violation(e):
                if self.meta.autoname == "hash":
                    # hash collision? try again
                    self.name = None
                    self.db_insert()
                    return

                frappe.msgprint(_("Duplicate name {0} {1}").format(self.doctype, self.name))
                raise frappe.DuplicateEntryError(self.doctype, self.name, e)

            elif frappe.db.is_unique_key_violation(e):
                # unique constraint
                self.show_unique_validation_message(e)

            else:
                raise

        self.set("__islocal", False)

    def db_update(self):
        if self.get("__islocal") or not self.name:
            self.db_insert()
            return

        d = self.get_valid_dict(convert_dates_to_str=True,
                                ignore_nulls=self.doctype in ('DocType', 'DocField', 'DocPerm'))

        # don't update name, as case might've been changed
        name = d['name']
        del d['name']

        columns = list(d)

        try:
            frappe.db.sql("""UPDATE `tab{doctype}`
				SET {values} WHERE `name`=%s""".format(
                doctype=self.doctype,
                values=", ".join(["`" + c + "`=%s" for c in columns])
            ), list(d.values()) + [name])
        except Exception as e:
            if frappe.db.is_unique_key_violation(e):
                self.show_unique_validation_message(e)
            else:
                raise

    def show_unique_validation_message(self, e):
        # TODO: Find a better way to extract fieldname
        if frappe.db.db_type != 'postgres':
            fieldname = str(e).split("'")[-2]
            label = None

            # unique_first_fieldname_second_fieldname is the constraint name
            # created using frappe.db.add_unique
            if "unique_" in fieldname:
                fieldname = fieldname.split("_", 1)[1]

            df = self.meta.get_field(fieldname)
            if df:
                label = df.label

            frappe.msgprint(_("{0} must be unique".format(label or fieldname)))

        # this is used to preserve traceback
        raise frappe.UniqueValidationError(self.doctype, self.name, e)

    def update_modified(self):
        '''Update modified timestamp'''
        self.set("modified", now())
        frappe.db.set_value(self.doctype, self.name, 'modified', self.modified, update_modified=False)

    def _fix_numeric_types(self):
        for df in self.meta.get("fields"):
            if df.fieldtype == "Check":
                self.set(df.fieldname, cint(self.get(df.fieldname)))

            elif self.get(df.fieldname) is not None:
                if df.fieldtype == "Int":
                    self.set(df.fieldname, cint(self.get(df.fieldname)))

                elif df.fieldtype in ("Float", "Currency", "Percent"):
                    self.set(df.fieldname, flt(self.get(df.fieldname)))

        if self.docstatus is not None:
            self.docstatus = cint(self.docstatus)

    def _get_missing_mandatory_fields(self):
        """Get mandatory fields that do not have any values"""

        def get_msg(df):
            if df.fieldtype in table_fields:
                return "{}: {}: {}".format(_("Error"), _("Data missing in table"), _(df.label))

            elif self.parentfield:
                return "{}: {} {} #{}: {}: {}".format(_("Error"), frappe.bold(_(self.doctype)),
                                                      _("Row"), self.idx, _("Value missing for"), _(df.label))

            else:
                return _("Error: Value missing for {0}: {1}").format(_(df.parent), _(df.label))

        missing = []

        for df in self.meta.get("fields", {"reqd": ('=', 1)}):
            if self.get(df.fieldname) in (None, []):
                missing.append((df.fieldname, get_msg(df)))

        # check for missing parent and parenttype
        if self.meta.istable:
            for fieldname in ("parent", "parenttype"):
                if not self.get(fieldname):
                    missing.append((fieldname, get_msg(frappe._dict(label=fieldname))))

        return missing

    def get_invalid_links(self, is_submittable=False):
        '''Returns list of invalid links and also updates fetch values if not set'''

        def get_msg(df, docname):
            if self.parentfield:
                return "{} #{}: {}: {}".format(_("Row"), self.idx, _(df.label), docname)
            else:
                return "{}: {}".format(_(df.label), docname)

        invalid_links = []
        cancelled_links = []

        for df in (self.meta.get_link_fields()
                   + self.meta.get("fields", {"fieldtype": ('=', "Dynamic Link")})):
            docname = self.get(df.fieldname)

            if docname:
                if df.fieldtype == "Link":
                    doctype = df.options
                    if not doctype:
                        frappe.throw(_("Options not set for link field {0}").format(df.fieldname))
                else:
                    doctype = self.get(df.options)
                    if not doctype:
                        frappe.throw(_("{0} must be set first").format(self.meta.get_label(df.options)))

                # MySQL is case insensitive. Preserve case of the original docname in the Link Field.

                # get a map of values ot fetch along with this link query
                # that are mapped as link_fieldname.source_fieldname in Options of
                # Readonly or Data or Text type fields

                fields_to_fetch = [
                    _df for _df in self.meta.get_fields_to_fetch(df.fieldname)
                    if
                    not _df.get('fetch_if_empty')
                    or (_df.get('fetch_if_empty') and not self.get(_df.fieldname))
                ]

                if not fields_to_fetch:
                    # cache a single value type
                    values = frappe._dict(name=frappe.db.get_value(doctype, docname,
                                                                   'name', cache=True))
                else:
                    values_to_fetch = ['name'] + [_df.fetch_from.split('.')[-1]
                                                  for _df in fields_to_fetch]

                    # don't cache if fetching other values too
                    values = frappe.db.get_value(doctype, docname,
                                                 values_to_fetch, as_dict=True)

                if frappe.get_meta(doctype).issingle:
                    values.name = doctype

                if values:
                    setattr(self, df.fieldname, values.name)

                    for _df in fields_to_fetch:
                        if self.is_new() or self.docstatus != 1 or _df.allow_on_submit:
                            setattr(self, _df.fieldname, values[_df.fetch_from.split('.')[-1]])

                    notify_link_count(doctype, docname)

                    if not values.name:
                        invalid_links.append((df.fieldname, docname, get_msg(df, docname)))

                    elif (df.fieldname != "amended_from"
                          and (is_submittable or self.meta.is_submittable) and frappe.get_meta(doctype).is_submittable
                          and cint(frappe.db.get_value(doctype, docname, "docstatus")) == 2):

                        cancelled_links.append((df.fieldname, docname, get_msg(df, docname)))

        return invalid_links, cancelled_links

    def _validate_selects(self):
        if frappe.flags.in_import:
            return

        for df in self.meta.get_select_fields():
            if df.fieldname == "naming_series" or not (self.get(df.fieldname) and df.options):
                continue

            options = (df.options or "").split("\n")

            # if only empty options
            if not filter(None, options):
                continue

            # strip and set
            self.set(df.fieldname, cstr(self.get(df.fieldname)).strip())
            value = self.get(df.fieldname)

            if value not in options and not (frappe.flags.in_test and value.startswith("_T-")):
                # show an elaborate message
                prefix = _("Row #{0}:").format(self.idx) if self.get("parentfield") else ""
                label = _(self.meta.get_label(df.fieldname))
                comma_options = '", "'.join(_(each) for each in options)

                frappe.throw(_('{0} {1} cannot be "{2}". It should be one of "{3}"').format(prefix, label,
                                                                                            value, comma_options))

    def _validate_constants(self):
        if frappe.flags.in_import or self.is_new() or self.flags.ignore_validate_constants:
            return

        constants = [d.fieldname for d in self.meta.get("fields", {"set_only_once": ('=', 1)})]
        if constants:
            values = frappe.db.get_value(self.doctype, self.name, constants, as_dict=True)

        for fieldname in constants:
            df = self.meta.get_field(fieldname)

            # This conversion to string only when fieldtype is Date
            if df.fieldtype == 'Date' or df.fieldtype == 'Datetime':
                value = str(values.get(fieldname))

            else:
                value = values.get(fieldname)

            if self.get(fieldname) != value:
                frappe.throw(_("Value cannot be changed for {0}").format(self.meta.get_label(fieldname)),
                             frappe.CannotChangeConstantError)

    def _validate_length(self):
        if frappe.flags.in_install:
            return

        if self.meta.issingle:
            # single doctype value type is mediumtext
            return

        type_map = frappe.db.type_map

        for fieldname, value in iteritems(self.get_valid_dict()):
            df = self.meta.get_field(fieldname)

            if not df or df.fieldtype == 'Check':
                # skip standard fields and Check fields
                continue

            column_type = type_map[df.fieldtype][0] or None

            if column_type == 'varchar':
                default_column_max_length = type_map[df.fieldtype][1] or None
                max_length = cint(df.get("length")) or cint(default_column_max_length)

                if len(cstr(value)) > max_length:
                    self.throw_length_exceeded_error(df, max_length, value)

            elif column_type in ('int', 'bigint', 'smallint'):
                max_length = max_positive_value[column_type]

                if abs(cint(value)) > max_length:
                    self.throw_length_exceeded_error(df, max_length, value)

    def throw_length_exceeded_error(self, df, max_length, value):
        if self.parentfield and self.idx:
            reference = _("{0}, Row {1}").format(_(self.doctype), self.idx)

        else:
            reference = "{0} {1}".format(_(self.doctype), self.name)

        frappe.throw(_("{0}: '{1}' ({3}) will get truncated, as max characters allowed is {2}") \
                     .format(reference, _(df.label), max_length, value), frappe.CharacterLengthExceededError,
                     title=_('Value too big'))

    def _validate_update_after_submit(self):
        # get the full doc with children
        db_values = frappe.get_doc(self.doctype, self.name).as_dict()

        for key in self.as_dict():
            df = self.meta.get_field(key)
            db_value = db_values.get(key)

            if df and not df.allow_on_submit and (self.get(key) or db_value):
                if df.fieldtype in table_fields:
                    # just check if the table size has changed
                    # individual fields will be checked in the loop for children
                    self_value = len(self.get(key))
                    db_value = len(db_value)

                else:
                    self_value = self.get_value(key)

                if self_value != db_value:
                    frappe.throw(_("Not allowed to change {0} after submission").format(df.label),
                                 frappe.UpdateAfterSubmitError)

    def _sanitize_content(self):
        """Sanitize HTML and Email in field values. Used to prevent XSS.

            - Ignore if 'Ignore XSS Filter' is checked or fieldtype is 'Code'
        """
        if frappe.flags.in_install:
            return

        for fieldname, value in self.get_valid_dict().items():
            if not value or not isinstance(value, string_types):
                continue

            value = frappe.as_unicode(value)

            if (u"<" not in value and u">" not in value):
                # doesn't look like html so no need
                continue

            elif "<!-- markdown -->" in value and not ("<script" in value or "javascript:" in value):
                # should be handled separately via the markdown converter function
                continue

            df = self.meta.get_field(fieldname)
            sanitized_value = value

            if df and df.get("fieldtype") in ("Data", "Code", "Small Text") and df.get("options") == "Email":
                sanitized_value = value

            elif df and (df.get("ignore_xss_filter")
                         or (df.get("fieldtype") == "Code" and df.get("options") != "Email")
                         or df.get("fieldtype") in ("Attach", "Attach Image", "Barcode")

                         # cancelled and submit but not update after submit should be ignored
                         or self.docstatus == 2
                         or (self.docstatus == 1 and not df.get("allow_on_submit"))):
                continue

            else:
                sanitized_value = value

            self.set(fieldname, sanitized_value)

    def _save_passwords(self):
        '''Save password field values in __Auth table'''
        if self.flags.ignore_save_passwords is True:
            return

        for df in self.meta.get('fields', {'fieldtype': ('=', 'Password')}):
            if self.flags.ignore_save_passwords and df.fieldname in self.flags.ignore_save_passwords: continue
            new_password = self.get(df.fieldname)
            if new_password and not self.is_dummy_password(new_password):
                # is not a dummy password like '*****'

                # set dummy password like '*****'
                self.set(df.fieldname, '*' * len(new_password))

    def get_password(self, fieldname='password', raise_exception=True):
        if self.get(fieldname) and not self.is_dummy_password(self.get(fieldname)):
            return self.get(fieldname)

        return "frappe"

    def is_dummy_password(self, pwd):
        return ''.join(set(pwd)) == '*'

    def precision(self, fieldname, parentfield=None):
        """Returns float precision for a particular field (or get global default).

        :param fieldname: Fieldname for which precision is required.
        :param parentfield: If fieldname is in child table."""
        from frappe.model.meta import get_field_precision

        if parentfield and not isinstance(parentfield, string_types):
            parentfield = parentfield.parentfield

        cache_key = parentfield or "main"

        if not hasattr(self, "_precision"):
            self._precision = frappe._dict()

        if cache_key not in self._precision:
            self._precision[cache_key] = frappe._dict()

        if fieldname not in self._precision[cache_key]:
            self._precision[cache_key][fieldname] = None

            doctype = self.meta.get_field(parentfield).options if parentfield else self.doctype
            df = frappe.get_meta(doctype).get_field(fieldname)

            if df.fieldtype in ("Currency", "Float", "Percent"):
                self._precision[cache_key][fieldname] = get_field_precision(df, self)

        return self._precision[cache_key][fieldname]

    def get_formatted(self, fieldname, doc=None, currency=None, absolute_value=False, translated=False):
        from frappe.utils.formatters import format_value

        df = self.meta.get_field(fieldname)
        if not df and fieldname in default_fields:
            from frappe.model.meta import get_default_df
            df = get_default_df(fieldname)

        val = self.get(fieldname)

        if translated:
            val = _(val)

        if absolute_value and isinstance(val, (int, float)):
            val = abs(self.get(fieldname))

        if not doc:
            doc = getattr(self, "parent_doc", None) or self

        return format_value(val, df=df, doc=doc, currency=currency)

    def is_print_hide(self, fieldname, df=None, for_print=True):
        """Returns true if fieldname is to be hidden for print.

        Print Hide can be set via the Print Format Builder or in the controller as a list
        of hidden fields. Example

            class MyDoc(Document):
                def __setup__(self):
                    self.print_hide = ["field1", "field2"]

        :param fieldname: Fieldname to be checked if hidden.
        """
        meta_df = self.meta.get_field(fieldname)
        if meta_df and meta_df.get("__print_hide"):
            return True

        print_hide = 0

        if self.get(fieldname) == 0 and not self.meta.istable:
            print_hide = (df and df.print_hide_if_no_value) or (meta_df and meta_df.print_hide_if_no_value)

        if not print_hide:
            if df and df.print_hide is not None:
                print_hide = df.print_hide
            elif meta_df:
                print_hide = meta_df.print_hide

        return print_hide

    def in_format_data(self, fieldname):
        """Returns True if shown via Print Format::`format_data` property.
            Called from within standard print format."""
        doc = getattr(self, "parent_doc", self)

        if hasattr(doc, "format_data_map"):
            return fieldname in doc.format_data_map
        else:
            return True

    def reset_values_if_no_permlevel_access(self, has_access_to, high_permlevel_fields):
        """If the user does not have permissions at permlevel > 0, then reset the values to original / default"""
        to_reset = []

        for df in high_permlevel_fields:
            if df.permlevel not in has_access_to and df.fieldtype not in display_fieldtypes:
                to_reset.append(df)

        if to_reset:
            if self.is_new():
                # if new, set default value
                ref_doc = frappe.new_doc(self.doctype)
            else:
                # get values from old doc
                if self.get('parent_doc'):
                    self.parent_doc.get_latest()
                    ref_doc = [d for d in self.parent_doc.get(self.parentfield) if d.name == self.name][0]
                else:
                    ref_doc = self.get_latest()

            for df in to_reset:
                self.set(df.fieldname, ref_doc.get(df.fieldname))

    def get_value(self, fieldname):
        df = self.meta.get_field(fieldname)
        val = self.get(fieldname)

        return self.cast(val, df)

    def cast(self, value, df):
        return cast_fieldtype(df.fieldtype, value)

    def _extract_images_from_text_editor(self):
        pass


def _filter(data, filters, limit=None):
    """pass filters as:
        {"key": "val", "key": ["!=", "val"],
        "key": ["in", "val"], "key": ["not in", "val"], "key": "^val",
        "key" : True (exists), "key": False (does not exist) }"""

    out, _filters = [], {}

    if not data:
        return out

    # setup filters as tuples
    if filters:
        for f in filters:
            fval = filters[f]

            if not isinstance(fval, (tuple, list)):
                if fval is True:
                    fval = ("not None", fval)
                elif fval is False:
                    fval = ("None", fval)
                elif isinstance(fval, string_types) and fval.startswith("^"):
                    fval = ("^", fval[1:])
                else:
                    fval = ("=", fval)

            _filters[f] = fval

    for d in data:
        add = True
        for f, fval in iteritems(_filters):
            if not frappe.compare(getattr(d, f, None), fval[0], fval[1]):
                add = False
                break

        if add:
            out.append(d)
            if limit and (len(out) - 1) == limit:
                break

    return out



# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# MIT License. See license.txt

from __future__ import unicode_literals

from six import iteritems, string_types

"""build query for doclistview and return results"""

import frappe.defaults
from frappe import _
from datetime import datetime
import frappe, json, copy, re
from frappe.model import optional_fields
from frappe.model.utils.user_settings import get_user_settings, update_user_settings
from frappe.utils import flt, cint, get_time, make_filter_tuple, get_filter, add_to_date, cstr, nowdate

class DatabaseQuery(object):
	def __init__(self, doctype, user=None):
		self.doctype = doctype
		self.tables = []
		self.conditions = []
		self.or_conditions = []
		self.fields = None
		self.user = user
		self.ignore_ifnull = False
		self.flags = frappe._dict()
		self.reference_doctype = None

	def execute(self, query=None, fields=None, filters=None, or_filters=None,
				docstatus=None, group_by=None, order_by=None, limit_start=False,
				limit_page_length=None, as_list=False, with_childnames=False, debug=False,
				ignore_permissions=False, user=None, with_comment_count=False,
				join='left join', distinct=False, start=None, page_length=None, limit=None,
				ignore_ifnull=False, save_user_settings=False, save_user_settings_fields=False,
				update=None, add_total_row=None, user_settings=None, reference_doctype=None, return_query=False, strict=True):
		if not ignore_permissions and not frappe.has_permission(self.doctype, "read", user=user):
			frappe.flags.error_message = _('Insufficient Permission for {0}').format(frappe.bold(self.doctype))
			raise frappe.PermissionError(self.doctype)

		# filters and fields swappable
		# its hard to remember what comes first
		if (isinstance(fields, dict)
				or (isinstance(fields, list) and fields and isinstance(fields[0], list))):
			# if fields is given as dict/list of list, its probably filters
			filters, fields = fields, filters

		elif fields and isinstance(filters, list) \
				and len(filters) > 1 and isinstance(filters[0], string_types):
			# if `filters` is a list of strings, its probably fields
			filters, fields = fields, filters

		if fields:
			self.fields = fields
		else:
			self.fields =  ["`tab{0}`.`name`".format(self.doctype)]

		if start: limit_start = start
		if page_length: limit_page_length = page_length
		if limit: limit_page_length = limit

		self.filters = filters or []
		self.or_filters = or_filters or []
		self.docstatus = docstatus or []
		self.group_by = group_by
		self.order_by = order_by
		self.limit_start = 0 if (limit_start is False) else cint(limit_start)
		self.limit_page_length = cint(limit_page_length) if limit_page_length else None
		self.with_childnames = with_childnames
		self.debug = debug
		self.join = join
		self.distinct = distinct
		self.as_list = as_list
		self.ignore_ifnull = ignore_ifnull
		self.flags.ignore_permissions = ignore_permissions
		self.user = user
		self.update = update
		self.user_settings_fields = copy.deepcopy(self.fields)
		self.return_query = return_query
		self.strict = strict

		# for contextual user permission check
		# to determine which user permission is applicable on link field of specific doctype
		self.reference_doctype = reference_doctype or self.doctype

		if user_settings:
			self.user_settings = json.loads(user_settings)

		if query:
			result = self.run_custom_query(query)
		else:
			result = self.build_and_run()
			if return_query:
				return result

		if with_comment_count and not as_list and self.doctype:
			self.add_comment_count(result)

		if save_user_settings:
			self.save_user_settings_fields = save_user_settings_fields
			self.update_user_settings()

		return result

	def build_and_run(self):
		args = self.prepare_args()
		args.limit = self.add_limit()

		if args.conditions:
			args.conditions = "where " + args.conditions

		if self.distinct:
			args.fields = 'distinct ' + args.fields
			args.order_by = '' # TODO: recheck for alternative

		query = """select %(fields)s
			from %(tables)s
			%(conditions)s
			%(group_by)s
			%(order_by)s
			%(limit)s""" % args

		if self.return_query:
			return query
		else:
			return frappe.db.sql(query, as_dict=not self.as_list, debug=self.debug, update=self.update)

	def prepare_args(self):
		self.parse_args()
		self.sanitize_fields()
		self.extract_tables()
		self.set_optional_columns()
		self.build_conditions()

		args = frappe._dict()

		if self.with_childnames:
			for t in self.tables:
				if t != "`tab" + self.doctype + "`":
					self.fields.append(t + ".name as '%s:name'" % t[4:-1])

		# query dict
		args.tables = self.tables[0]

		# left join parent, child tables
		for child in self.tables[1:]:
			args.tables += " {join} {child} on ({child}.parent = {main}.name)".format(join=self.join,
																					  child=child, main=self.tables[0])

		if self.grouped_or_conditions:
			self.conditions.append("({0})".format(" or ".join(self.grouped_or_conditions)))

		args.conditions = ' and '.join(self.conditions)

		if self.or_conditions:
			args.conditions += (' or ' if args.conditions else "") + \
							   ' or '.join(self.or_conditions)

		self.set_field_tables()

		args.fields = ', '.join(self.fields)

		self.set_order_by(args)

		self.validate_order_by_and_group_by(args.order_by)
		args.order_by = args.order_by and (" order by " + args.order_by) or ""

		self.validate_order_by_and_group_by(self.group_by)
		args.group_by = self.group_by and (" group by " + self.group_by) or ""

		return args

	def parse_args(self):
		"""Convert fields and filters from strings to list, dicts"""
		if isinstance(self.fields, string_types):
			if self.fields == "*":
				self.fields = ["*"]
			else:
				try:
					self.fields = json.loads(self.fields)
				except ValueError:
					self.fields = [f.strip() for f in self.fields.split(",")]

		# remove empty strings / nulls in fields
		self.fields = [f for f in self.fields if f]

		for filter_name in ["filters", "or_filters"]:
			filters = getattr(self, filter_name)
			if isinstance(filters, string_types):
				filters = json.loads(filters)

			if isinstance(filters, dict):
				fdict = filters
				filters = []
				for key, value in iteritems(fdict):
					filters.append(make_filter_tuple(self.doctype, key, value))
			setattr(self, filter_name, filters)

	def sanitize_fields(self):
		'''
			regex : ^.*[,();].*
			purpose : The regex will look for malicious patterns like `,`, '(', ')', ';' in each
					field which may leads to sql injection.
			example :
				field = "`DocType`.`issingle`, version()"
			As field contains `,` and mysql function `version()`, with the help of regex
			the system will filter out this field.
		'''

		sub_query_regex = re.compile("^.*[,();].*")
		blacklisted_keywords = ['select', 'create', 'insert', 'delete', 'drop', 'update', 'case']
		blacklisted_functions = ['concat', 'concat_ws', 'if', 'ifnull', 'nullif', 'coalesce',
								 'connection_id', 'current_user', 'database', 'last_insert_id', 'session_user',
								 'system_user', 'user', 'version']

		def _raise_exception():
			frappe.throw(_('Use of sub-query or function is restricted'), frappe.DataError)

		def _is_query(field):
			if re.compile(r"^(select|delete|update|drop|create)\s").match(field):
				_raise_exception()

			elif re.compile(r"\s*[0-9a-zA-z]*\s*( from | group by | order by | where | join )").match(field):
				_raise_exception()

		for field in self.fields:
			if sub_query_regex.match(field):
				if any(keyword in field.lower().split() for keyword in blacklisted_keywords):
					_raise_exception()

				if any("({0}".format(keyword) in field.lower() for keyword in blacklisted_keywords):
					_raise_exception()

				if any("{0}(".format(keyword) in field.lower() for keyword in blacklisted_functions):
					_raise_exception()

			if re.compile(r"[0-9a-zA-Z]+\s*'").match(field):
				_raise_exception()

			if re.compile(r"[0-9a-zA-Z]+\s*,").match(field):
				_raise_exception()

			_is_query(field)

			if self.strict:
				if re.compile(r".*/\*.*").match(field):
					frappe.throw(_('Illegal SQL Query'))

				if re.compile(r".*\s(union).*\s").match(field.lower()):
					frappe.throw(_('Illegal SQL Query'))

	def extract_tables(self):
		"""extract tables from fields"""
		self.tables = ['`tab' + self.doctype + '`']

		# add tables from fields
		if self.fields:
			for f in self.fields:
				if ( not ("tab" in f and "." in f) ) or ("locate(" in f) or ("strpos(" in f) or ("count(" in f):
					continue

				table_name = f.split('.')[0]
				if table_name.lower().startswith('group_concat('):
					table_name = table_name[13:]
				if table_name.lower().startswith('ifnull('):
					table_name = table_name[7:]
				if not table_name[0]=='`':
					table_name = '`' + table_name + '`'
				if not table_name in self.tables:
					self.append_table(table_name)

	def append_table(self, table_name):
		self.tables.append(table_name)
		doctype = table_name[4:-1]
		if (not self.flags.ignore_permissions) and (not frappe.has_permission(doctype)):
			frappe.flags.error_message = _('Insufficient Permission for {0}').format(frappe.bold(doctype))
			raise frappe.PermissionError(doctype)

	def set_field_tables(self):
		'''If there are more than one table, the fieldname must not be ambiguous.
		If the fieldname is not explicitly mentioned, set the default table'''
		if len(self.tables) > 1:
			for i, f in enumerate(self.fields):
				if '.' not in f:
					self.fields[i] = '{0}.{1}'.format(self.tables[0], f)

	def set_optional_columns(self):
		"""Removes optional columns like `_user_tags`, `_comments` etc. if not in table"""
		columns = frappe.db.get_table_columns(self.doctype)

		# remove from fields
		to_remove = []
		for fld in self.fields:
			for f in optional_fields:
				if f in fld and not f in columns:
					to_remove.append(fld)

		for fld in to_remove:
			del self.fields[self.fields.index(fld)]

		# remove from filters
		to_remove = []
		for each in self.filters:
			if isinstance(each, string_types):
				each = [each]

			for element in each:
				if element in optional_fields and element not in columns:
					to_remove.append(each)

		for each in to_remove:
			if isinstance(self.filters, dict):
				del self.filters[each]
			else:
				self.filters.remove(each)

	def build_conditions(self):
		self.conditions = []
		self.grouped_or_conditions = []
		self.build_filter_conditions(self.filters, self.conditions)
		self.build_filter_conditions(self.or_filters, self.grouped_or_conditions)

		# match conditions
		if not self.flags.ignore_permissions:
			match_conditions = self.build_match_conditions()
			if match_conditions:
				self.conditions.append("(" + match_conditions + ")")

	def build_filter_conditions(self, filters, conditions, ignore_permissions=None):
		"""build conditions from user filters"""
		if ignore_permissions is not None:
			self.flags.ignore_permissions = ignore_permissions

		if isinstance(filters, dict):
			filters = [filters]

		for f in filters:
			if isinstance(f, string_types):
				conditions.append(f)
			else:
				conditions.append(self.prepare_filter_condition(f))

	def prepare_filter_condition(self, f):
		"""Returns a filter condition in the format:
				ifnull(`tabDocType`.`fieldname`, fallback) operator "value"
		"""

		f = get_filter(self.doctype, f)

		tname = ('`tab' + f.doctype + '`')
		if not tname in self.tables:
			self.append_table(tname)

		if 'ifnull(' in f.fieldname:
			column_name = f.fieldname
		else:
			column_name = '{tname}.{fname}'.format(tname=tname,
												   fname=f.fieldname)

		can_be_null = True

		# prepare in condition
		if f.operator.lower() in ('ancestors of', 'descendants of', 'not ancestors of', 'not descendants of'):
			values = f.value or ''

			# TODO: handle list and tuple
			# if not isinstance(values, (list, tuple)):
			# 	values = values.split(",")

			ref_doctype = f.doctype

			if frappe.get_meta(f.doctype).get_field(f.fieldname) is not None :
				ref_doctype = frappe.get_meta(f.doctype).get_field(f.fieldname).options

			result=[]
			lft, rgt = frappe.db.get_value(ref_doctype, f.value, ["lft", "rgt"])

			# Get descendants elements of a DocType with a tree structure
			if f.operator.lower() in ('descendants of', 'not descendants of') :
				result = frappe.get_all(ref_doctype, filters={
					'lft': ['>', lft],
					'rgt': ['<', rgt]
				}, order_by='`lft` ASC')
			else :
				# Get ancestor elements of a DocType with a tree structure
				result = frappe.get_all(ref_doctype, filters={
					'lft': ['<', lft],
					'rgt': ['>', rgt]
				}, order_by='`lft` DESC')

			fallback = "''"
			value = [frappe.db.escape((v.name or '').strip(), percent=False) for v in result]
			if len(value):
				value = "({0})".format(", ".join(value))
			else:
				value = "('')"
			# changing operator to IN as the above code fetches all the parent / child values and convert into tuple
			# which can be directly used with IN operator to query.
			f.operator = 'not in' if f.operator.lower() in ('not ancestors of', 'not descendants of') else 'in'


		elif f.operator.lower() in ('in', 'not in'):
			values = f.value or ''
			if isinstance(values, frappe.string_types):
				values = values.split(",")

			fallback = "''"
			value = [frappe.db.escape((v or '').strip(), percent=False) for v in values]
			if len(value):
				value = "({0})".format(", ".join(value))
			else:
				value = "('')"
		else:
			df = frappe.get_meta(f.doctype).get("fields", {"fieldname": f.fieldname})
			df = df[0] if df else None

			if df and df.fieldtype in ("Check", "Float", "Int", "Currency", "Percent"):
				can_be_null = False

			if f.operator.lower() in ('previous', 'next'):
				if f.operator.lower() == "previous":
					if f.value == "1 week":
						date_range = [add_to_date(nowdate(), days=-7), nowdate()]
					elif f.value == "1 month":
						date_range = [add_to_date(nowdate(), months=-1), nowdate()]
					elif f.value == "3 months":
						date_range = [add_to_date(nowdate(), months=-3), nowdate()]
					elif f.value == "6 months":
						date_range = [add_to_date(nowdate(), months=-6), nowdate()]
					elif f.value == "1 year":
						date_range = [add_to_date(nowdate(), years=-1), nowdate()]
				elif f.operator.lower() == "next":
					if f.value == "1 week":
						date_range = [nowdate(), add_to_date(nowdate(), days=7)]
					elif f.value == "1 month":
						date_range = [nowdate(), add_to_date(nowdate(), months=1)]
					elif f.value == "3 months":
						date_range = [nowdate(), add_to_date(nowdate(), months=3)]
					elif f.value == "6 months":
						date_range = [nowdate(), add_to_date(nowdate(), months=6)]
					elif f.value == "1 year":
						date_range = [nowdate(), add_to_date(nowdate(), years=1)]
				f.operator = "Between"
				f.value = date_range
				fallback = "'0001-01-01 00:00:00'"

			if f.operator in ('>', '<') and (f.fieldname in ('creation', 'modified')):
				value = cstr(f.value)
				fallback = "NULL"

			elif f.operator.lower() in ('between') and \
					(f.fieldname in ('creation', 'modified') or (df and (df.fieldtype=="Date" or df.fieldtype=="Datetime"))):

				value = get_between_date_filter(f.value, df)
				fallback = "'0001-01-01 00:00:00'"

			elif df and df.fieldtype=="Date":
				value = frappe.db.format_date(f.value)
				fallback = "'0001-01-01'"

			elif (df and df.fieldtype=="Datetime") or isinstance(f.value, datetime):
				value = frappe.db.format_datetime(f.value)
				fallback = "'0001-01-01 00:00:00'"

			elif df and df.fieldtype=="Time":
				value = get_time(f.value).strftime("%H:%M:%S.%f")
				fallback = "'00:00:00'"

			elif f.operator.lower() == "is":
				if f.value == 'set':
					f.operator = '!='
				elif f.value == 'not set':
					f.operator = '='

				value = ""
				fallback = "''"
				can_be_null = True

				if 'ifnull' not in column_name:
					column_name = 'ifnull({}, {})'.format(column_name, fallback)

			elif f.operator.lower() in ("like", "not like") or (isinstance(f.value, string_types) and
																(not df or df.fieldtype not in ["Float", "Int", "Currency", "Percent", "Check"])):
				value = "" if f.value==None else f.value
				fallback = "''"

				if f.operator.lower() in ("like", "not like") and isinstance(value, string_types):
					# because "like" uses backslash (\) for escaping
					value = value.replace("\\", "\\\\").replace("%", "%%")

			elif f.operator == '=' and df and df.fieldtype in ['Link', 'Data']: # TODO: Refactor if possible
				value = f.value or "''"
				fallback = "''"

			else:
				value = flt(f.value)
				fallback = 0

			# escape value
			if isinstance(value, string_types) and not f.operator.lower() == 'between':
				value = "{0}".format(frappe.db.escape(value, percent=False))

		if (self.ignore_ifnull
				or not can_be_null
				or (f.value and f.operator.lower() in ('=', 'like'))
				or 'ifnull(' in column_name.lower()):
			condition = '{column_name} {operator} {value}'.format(
				column_name=column_name, operator=f.operator,
				value=value)
		else:
			condition = 'ifnull({column_name}, {fallback}) {operator} {value}'.format(
				column_name=column_name, fallback=fallback, operator=f.operator,
				value=value)

		return condition

	def build_match_conditions(self, as_condition=True):
		"""add match conditions if applicable"""
		self.match_filters = []
		self.match_conditions = []
		only_if_shared = False
		if not self.user:
			self.user = None

		if not self.tables: self.extract_tables()

		meta = frappe.get_meta(self.doctype)
		role_permissions = frappe.permissions.get_role_permissions(meta, user=self.user)
		self.shared = frappe.share.get_shared(self.doctype, self.user)

		if (not meta.istable and
				not role_permissions.get("read") and
				not self.flags.ignore_permissions and
				not has_any_user_permission_for_doctype(self.doctype, self.user, self.reference_doctype)):
			only_if_shared = True
			if not self.shared:
				frappe.throw(_("No permission to read {0}").format(self.doctype), frappe.PermissionError)
			else:
				self.conditions.append(self.get_share_condition())

		else:
			#if has if_owner permission skip user perm check
			if role_permissions.get("if_owner", {}).get("read"):
				self.match_conditions.append("`tab{0}`.`owner` = {1}".format(self.doctype,
																			 frappe.db.escape(self.user, percent=False)))
			# add user permission only if role has read perm
			elif role_permissions.get("read"):
				# get user permissions
				user_permissions = frappe.permissions.get_user_permissions(self.user)
				self.add_user_permissions(user_permissions)

		if as_condition:
			conditions = ""
			if self.match_conditions:
				# will turn out like ((blog_post in (..) and blogger in (...)) or (blog_category in (...)))
				conditions = "((" + ") or (".join(self.match_conditions) + "))"

			doctype_conditions = self.get_permission_query_conditions()
			if doctype_conditions:
				conditions += (' and ' + doctype_conditions) if conditions else doctype_conditions

			# share is an OR condition, if there is a role permission
			if not only_if_shared and self.shared and conditions:
				conditions =  "({conditions}) or ({shared_condition})".format(
					conditions=conditions, shared_condition=self.get_share_condition())

			return conditions

		else:
			return self.match_filters

	def get_share_condition(self):
		return """`tab{0}`.name in ({1})""".format(self.doctype, ", ".join(["%s"] * len(self.shared))) % \
			   tuple([frappe.db.escape(s, percent=False) for s in self.shared])

	def add_user_permissions(self, user_permissions):
		meta = frappe.get_meta(self.doctype)
		doctype_link_fields = []
		doctype_link_fields = meta.get_link_fields()

		# append current doctype with fieldname as 'name' as first link field
		doctype_link_fields.append(dict(
			options=self.doctype,
			fieldname='name',
		))

		match_filters = {}
		match_conditions = []
		for df in doctype_link_fields:
			if df.get('ignore_user_permissions'): continue

			user_permission_values = user_permissions.get(df.get('options'), {})

			if user_permission_values:
				docs = []
				if frappe.get_system_settings("apply_strict_user_permissions"):
					condition = ""
				else:
					empty_value_condition = "ifnull(`tab{doctype}`.`{fieldname}`, '')=''".format(
						doctype=self.doctype, fieldname=df.get('fieldname')
					)
					condition = empty_value_condition + " or "

				for permission in user_permission_values:
					if not permission.get('applicable_for'):
						docs.append(permission.get('doc'))

					# append docs based on user permission applicable on reference doctype

					# this is useful when getting list of docs from a link field

					# in this case parent doctype of the link
					# will be the reference doctype

					elif df.get('fieldname') == 'name' and self.reference_doctype:
						if permission.get('applicable_for') == self.reference_doctype:
							docs.append(permission.get('doc'))

					elif permission.get('applicable_for') == self.doctype:
						docs.append(permission.get('doc'))

				if docs:
					condition += "`tab{doctype}`.`{fieldname}` in ({values})".format(
						doctype=self.doctype,
						fieldname=df.get('fieldname'),
						values=", ".join(
							[(frappe.db.escape(doc, percent=False)) for doc in docs])
					)

					match_conditions.append("({condition})".format(condition=condition))
					match_filters[df.get('options')] = docs

		if match_conditions:
			self.match_conditions.append(" and ".join(match_conditions))

		if match_filters:
			self.match_filters.append(match_filters)

	def get_permission_query_conditions(self):
		condition_methods = frappe.get_hooks("permission_query_conditions", {}).get(self.doctype, [])
		if condition_methods:
			conditions = []
			for method in condition_methods:
				c = frappe.call(frappe.get_attr(method), self.user)
				if c:
					conditions.append(c)

			return " and ".join(conditions) if conditions else None

	def run_custom_query(self, query):
		if '%(key)s' in query:
			query = query.replace('%(key)s', '`name`')
		return frappe.db.sql(query, as_dict = (not self.as_list))

	def set_order_by(self, args):
		meta = frappe.get_meta(self.doctype)

		if self.order_by:
			args.order_by = self.order_by
		else:
			args.order_by = ""

			# don't add order by from meta if a mysql group function is used without group by clause
			group_function_without_group_by = (len(self.fields)==1 and
											   (	self.fields[0].lower().startswith("count(")
													or self.fields[0].lower().startswith("min(")
													or self.fields[0].lower().startswith("max(")
													) and not self.group_by)

			if not group_function_without_group_by:
				sort_field = sort_order = None
				if meta.sort_field and ',' in meta.sort_field:
					# multiple sort given in doctype definition
					# Example:
					# `idx desc, modified desc`
					# will covert to
					# `tabItem`.`idx` desc, `tabItem`.`modified` desc
					args.order_by = ', '.join(['`tab{0}`.`{1}` {2}'.format(self.doctype,
																		   f.split()[0].strip(), f.split()[1].strip()) for f in meta.sort_field.split(',')])
				else:
					sort_field = meta.sort_field or 'modified'
					sort_order = (meta.sort_field and meta.sort_order) or 'desc'

					args.order_by = "`tab{0}`.`{1}` {2}".format(self.doctype, sort_field or "modified", sort_order or "desc")

				# draft docs always on top
				if meta.is_submittable:
					args.order_by = "`tab{0}`.docstatus asc, {1}".format(self.doctype, args.order_by)

	def validate_order_by_and_group_by(self, parameters):
		"""Check order by, group by so that atleast one column is selected and does not have subquery"""
		if not parameters:
			return

		_lower = parameters.lower()
		if 'select' in _lower and ' from ' in _lower:
			frappe.throw(_('Cannot use sub-query in order by'))

		if re.compile(r".*[^a-z0-9-_ ,`'\"\.\(\)].*").match(_lower):
			frappe.throw(_('Illegal SQL Query'))

		for field in parameters.split(","):
			if "." in field and field.strip().startswith("`tab"):
				tbl = field.strip().split('.')[0]
				if tbl not in self.tables:
					if tbl.startswith('`'):
						tbl = tbl[4:-1]
					frappe.throw(_("Please select atleast 1 column from {0} to sort/group").format(tbl))

	def add_limit(self):
		if self.limit_page_length:
			return 'limit %s offset %s' % (self.limit_page_length, self.limit_start)
		else:
			return ''

	def add_comment_count(self, result):
		for r in result:
			if not r.name:
				continue

			r._comment_count = 0
			if "_comments" in r:
				r._comment_count = len(json.loads(r._comments or "[]"))

	def update_user_settings(self):
		# update user settings if new search
		user_settings = json.loads(get_user_settings(self.doctype))

		if hasattr(self, 'user_settings'):
			user_settings.update(self.user_settings)

		if self.save_user_settings_fields:
			user_settings['fields'] = self.user_settings_fields

		update_user_settings(self.doctype, user_settings)

def get_order_by(doctype, meta):
	order_by = ""

	sort_field = sort_order = None
	if meta.sort_field and ',' in meta.sort_field:
		# multiple sort given in doctype definition
		# Example:
		# `idx desc, modified desc`
		# will covert to
		# `tabItem`.`idx` desc, `tabItem`.`modified` desc
		order_by = ', '.join(['`tab{0}`.`{1}` {2}'.format(doctype,
														  f.split()[0].strip(), f.split()[1].strip()) for f in meta.sort_field.split(',')])
	else:
		sort_field = meta.sort_field or 'modified'
		sort_order = (meta.sort_field and meta.sort_order) or 'desc'

		order_by = "`tab{0}`.`{1}` {2}".format(doctype, sort_field or "modified", sort_order or "desc")

	# draft docs always on top
	if meta.is_submittable:
		order_by = "`tab{0}`.docstatus asc, {1}".format(doctype, order_by)

	return order_by



def get_list(doctype, *args, **kwargs):
	'''wrapper for DatabaseQuery'''
	kwargs.pop('cmd', None)
	kwargs.pop('ignore_permissions', None)
	kwargs.pop('data', None)
	kwargs.pop('strict', None)

	# If doctype is child table
	if frappe.is_table(doctype):
		# Example frappe.db.get_list('Purchase Receipt Item', {'parent': 'Purchase Receipt'})
		# Here purchase receipt is the parent doctype of the child doctype Purchase Receipt Item

		if not kwargs.get('parent'):
			frappe.flags.error_message = _('Parent is required to get child table data')
			raise frappe.PermissionError(doctype)

		# check_parent_permission(kwargs.get('parent'), doctype)
		del kwargs['parent']

	return DatabaseQuery(doctype).execute(None, *args, **kwargs)

def is_parent_only_filter(doctype, filters):
	#check if filters contains only parent doctype
	only_parent_doctype = True

	if isinstance(filters, list):
		for flt in filters:
			if doctype not in flt:
				only_parent_doctype = False
			if 'Between' in flt:
				flt[3] = get_between_date_filter(flt[3])

	return only_parent_doctype

def has_any_user_permission_for_doctype(doctype, user, applicable_for):
	user_permissions = dict([])
	doctype_user_permissions = user_permissions.get(doctype, [])

	for permission in doctype_user_permissions:
		if not permission.applicable_for or permission.applicable_for == applicable_for:
			return True

	return False

def get_between_date_filter(value, df=None):
	'''
		return the formattted date as per the given example
		[u'2017-11-01', u'2017-11-03'] => '2017-11-01 00:00:00.000000' AND '2017-11-04 00:00:00.000000'
	'''
	from_date = frappe.utils.nowdate()
	to_date = frappe.utils.nowdate()

	if value and isinstance(value, (list, tuple)):
		if len(value) >= 1: from_date = value[0]
		if len(value) >= 2: to_date = value[1]

	if not df or (df and df.fieldtype == 'Datetime'):
		to_date = add_to_date(to_date, days=1)

	if df and df.fieldtype == 'Datetime':
		data = "'%s' AND '%s'" % (
			frappe.db.format_datetime(from_date),
			frappe.db.format_datetime(to_date))
	else:
		data = "'%s' AND '%s'" % (
			frappe.db.format_date(from_date),
			frappe.db.format_date(to_date))

	return data